"""
Regression Model
----------------
A `RegressionModel` forecasts future values of a target series based on

* The target series (past lags only)

* An optional past_covariates series (past lags only)

* An optional future_covariates series (possibly past and future lags)

* Available static covariates


The regression models are learned in a supervised way, and they can wrap around any "scikit-learn like" regression model
acting on tabular data having ``fit()`` and ``predict()`` methods.

Darts also provides :class:`LinearRegressionModel` and :class:`RandomForest`, which are regression models
wrapping around scikit-learn linear regression and random forest regression, respectively.

Behind the scenes this model is tabularizing the time series data to make it work with regression models.

The lags can be specified either using an integer - in which case it represents the _number_ of (past or future) lags
to take into consideration, or as a list - in which case the lags have to be enumerated (strictly negative values
denoting past lags and positive values including 0 denoting future lags).
When static covariates are present, they are appended to the lagged features. When multiple time series are passed,
if their static covariates do not have the same size, the shorter ones are padded with 0 valued features.
"""
from collections import OrderedDict
from typing import Any, List, Optional, Sequence, Tuple, Union

import numpy as np
from sklearn.linear_model import LinearRegression

from darts.logging import get_logger, raise_if, raise_if_not, raise_log
from darts.models.forecasting.forecasting_model import GlobalForecastingModel
from darts.timeseries import TimeSeries
from darts.utils.data.tabularization import (
    add_static_covariates_to_lagged_data,
    create_lagged_component_names,
    create_lagged_training_data,
)
from darts.utils.multioutput import MultiOutputRegressor
from darts.utils.utils import (
    _check_quantiles,
    get_single_series,
    seq2series,
    series2seq,
)

logger = get_logger(__name__)

try:
    from catboost import CatBoostRegressor
except ModuleNotFoundError:
    logger.warning(
        "The catboost module could not be imported. "
        "To enable support for CatBoostRegressor, "
        "follow the instruction in the README: "
        "https://github.com/unit8co/darts/blob/master/INSTALL.md"
    )


class RegressionModel(GlobalForecastingModel):
    def __init__(
        self,
        lags: Union[int, list] = None,
        lags_past_covariates: Union[int, List[int]] = None,
        lags_future_covariates: Union[Tuple[int, int], List[int]] = None,
        output_chunk_length: int = 1,
        add_encoders: Optional[dict] = None,
        model=None,
        multi_models: Optional[bool] = True,
        use_static_covariates: bool = True,
    ):
        """Regression Model
        Can be used to fit any scikit-learn-like regressor class to predict the target time series from lagged values.

        Parameters
        ----------
        lags
            Lagged target values used to predict the next time step. If an integer is given the last `lags` past lags
            are used (from -1 backward). Otherwise, a list of integers with lags is required (each lag must be < 0).
        lags_past_covariates
            Number of lagged past_covariates values used to predict the next time step. If an integer is given the last
            `lags_past_covariates` past lags are used (inclusive, starting from lag -1). Otherwise a list of integers
            with lags < 0 is required.
        lags_future_covariates
            Number of lagged future_covariates values used to predict the next time step. If a tuple (past, future) is
            given the last `past` lags in the past are used (inclusive, starting from lag -1) along with the first
            `future` future lags (starting from 0 - the prediction time - up to `future - 1` included). Otherwise a list
            of integers with lags is required.
        output_chunk_length
            Number of time steps predicted at once by the internal regression model. Does not have to equal the forecast
            horizon `n` used in `predict()`. However, setting `output_chunk_length` equal to the forecast horizon may
            be useful if the covariates don't extend far enough into the future.
        add_encoders
            A large number of past and future covariates can be automatically generated with `add_encoders`.
            This can be done by adding multiple pre-defined index encoders and/or custom user-made functions that
            will be used as index encoders. Additionally, a transformer such as Darts' :class:`Scaler` can be added to
            transform the generated covariates. This happens all under one hood and only needs to be specified at
            model creation.
            Read :meth:`SequentialEncoder <darts.dataprocessing.encoders.SequentialEncoder>` to find out more about
            ``add_encoders``. Default: ``None``. An example showing some of ``add_encoders`` features:

            .. highlight:: python
            .. code-block:: python

                add_encoders={
                    'cyclic': {'future': ['month']},
                    'datetime_attribute': {'future': ['hour', 'dayofweek']},
                    'position': {'past': ['relative'], 'future': ['relative']},
                    'custom': {'past': [lambda idx: (idx.year - 1950) / 50]},
                    'transformer': Scaler()
                }
            ..
        model
            Scikit-learn-like model with ``fit()`` and ``predict()`` methods. Also possible to use model that doesn't
            support multi-output regression for multivariate timeseries, in which case one regressor
            will be used per component in the multivariate series.
            If None, defaults to: ``sklearn.linear_model.LinearRegression(n_jobs=-1)``.
        multi_models
            If True, a separate model will be trained for each future lag to predict. If False, a single model is
            trained to predict at step 'output_chunk_length' in the future. Default: True.
        use_static_covariates
            Whether the model should use static covariate information in case the input `series` passed to ``fit()``
            contain static covariates. If ``True``, and static covariates are available at fitting time, will enforce
            that all target `series` have the same static covariate dimensionality in ``fit()`` and ``predict()`.
        """

        super().__init__(add_encoders=add_encoders)

        self.model = model
        self.lags = {}
        self.output_chunk_length = None
        self.input_dim = None
        self.multi_models = multi_models
        self._considers_static_covariates = use_static_covariates
        self._static_covariates_shape: Optional[Tuple[int, int]] = None
        self._lagged_feature_names: Optional[List[str]] = None

        # model checks
        if self.model is None:
            self.model = LinearRegression(n_jobs=-1)

        if not callable(getattr(self.model, "fit", None)):
            raise_log(
                Exception("Provided model object must have a fit() method", logger)
            )
        if not callable(getattr(self.model, "predict", None)):
            raise_log(
                Exception("Provided model object must have a predict() method", logger)
            )

        # check lags
        raise_if(
            (lags is None)
            and (lags_future_covariates is None)
            and (lags_past_covariates is None),
            "At least one of `lags`, `lags_future_covariates` or `lags_past_covariates` must be not None.",
        )

        lags_type_checks = [
            (lags, "lags"),
            (lags_past_covariates, "lags_past_covariates"),
        ]

        for _lags, lags_name in lags_type_checks:
            raise_if_not(
                isinstance(_lags, (int, list)) or _lags is None,
                f"`{lags_name}` must be of type int or list. Given: {type(_lags)}.",
            )
            raise_if(
                isinstance(_lags, bool),
                f"`{lags_name}` must be of type int or list, not bool.",
            )

        raise_if_not(
            isinstance(lags_future_covariates, (tuple, list))
            or lags_future_covariates is None,
            f"`lags_future_covariates` must be of type tuple or list. Given: {type(lags_future_covariates)}.",
        )

        if isinstance(lags_future_covariates, tuple):
            raise_if_not(
                len(lags_future_covariates) == 2
                and isinstance(lags_future_covariates[0], int)
                and isinstance(lags_future_covariates[1], int),
                "`lags_future_covariates` tuple must be of length 2, and must contain two integers",
            )
            raise_if(
                isinstance(lags_future_covariates[0], bool)
                or isinstance(lags_future_covariates[1], bool),
                "`lags_future_covariates` tuple must contain integers, not bool",
            )

        # set lags
        if isinstance(lags, int):
            raise_if_not(lags > 0, f"`lags` must be strictly positive. Given: {lags}.")
            # selecting last `lags` lags, starting from position 1 (skipping current, pos 0, the one we want to predict)
            self.lags["target"] = list(range(-lags, 0))
        elif isinstance(lags, list):
            for lag in lags:
                raise_if(
                    not isinstance(lag, int) or (lag >= 0),
                    f"Every element of `lags` must be a strictly negative integer. Given: {lags}.",
                )
            if lags:
                self.lags["target"] = sorted(lags)

        if isinstance(lags_past_covariates, int):
            raise_if_not(
                lags_past_covariates > 0,
                f"`lags_past_covariates` must be an integer > 0. Given: {lags_past_covariates}.",
            )
            self.lags["past"] = list(range(-lags_past_covariates, 0))
        elif isinstance(lags_past_covariates, list):
            for lag in lags_past_covariates:
                raise_if(
                    not isinstance(lag, int) or (lag >= 0),
                    f"Every element of `lags_covariates` must be an integer < 0. Given: {lags_past_covariates}.",
                )
            if lags_past_covariates:
                self.lags["past"] = sorted(lags_past_covariates)

        if isinstance(lags_future_covariates, tuple):
            raise_if_not(
                lags_future_covariates[0] >= 0 and lags_future_covariates[1] >= 0,
                f"`lags_future_covariates` tuple must contain integers >= 0. Given: {lags_future_covariates}.",
            )
            if (
                lags_future_covariates[0] is not None
                and lags_future_covariates[1] is not None
            ):
                if not (
                    lags_future_covariates[0] == 0 and lags_future_covariates[1] == 0
                ):
                    self.lags["future"] = list(
                        range(-lags_future_covariates[0], lags_future_covariates[1])
                    )
        elif isinstance(lags_future_covariates, list):
            for lag in lags_future_covariates:
                raise_if(
                    not isinstance(lag, int) or isinstance(lag, bool),
                    f"Every element of `lags_future_covariates` must be an integer. Given: {lags_future_covariates}.",
                )
            if lags_future_covariates:
                self.lags["future"] = sorted(lags_future_covariates)

        # check and set output_chunk_length
        raise_if_not(
            isinstance(output_chunk_length, int) and output_chunk_length > 0,
            f"output_chunk_length must be an integer greater than 0. Given: {output_chunk_length}",
        )
        self.output_chunk_length = output_chunk_length

        self.pred_dim = self.output_chunk_length if self.multi_models else 1

    @property
    def _model_encoder_settings(
        self,
    ) -> Tuple[int, int, bool, bool, Optional[List[int]], Optional[List[int]]]:
        target_lags = self.lags.get("target", [0])
        lags_past_covariates = self.lags.get("past", None)
        if lags_past_covariates is not None:
            lags_past_covariates = [
                min(lags_past_covariates)
                - int(not self.multi_models) * (self.output_chunk_length - 1),
                max(lags_past_covariates),
            ]
        lags_future_covariates = self.lags.get("future", None)
        if lags_future_covariates is not None:
            lags_future_covariates = [
                min(lags_future_covariates)
                - int(not self.multi_models) * (self.output_chunk_length - 1),
                max(lags_future_covariates),
            ]
        return (
            abs(min(target_lags)),
            self.output_chunk_length,
            lags_past_covariates is not None,
            lags_future_covariates is not None,
            lags_past_covariates,
            lags_future_covariates,
        )

    @property
    def extreme_lags(
        self,
    ) -> Tuple[
        Optional[int],
        Optional[int],
        Optional[int],
        Optional[int],
        Optional[int],
        Optional[int],
    ]:
        min_target_lag = self.lags.get("target")[0] if "target" in self.lags else None
        max_target_lag = self.output_chunk_length - 1
        min_past_cov_lag = self.lags.get("past")[0] if "past" in self.lags else None
        max_past_cov_lag = self.lags.get("past")[-1] if "past" in self.lags else None
        min_future_cov_lag = (
            self.lags.get("future")[0] if "future" in self.lags else None
        )
        max_future_cov_lag = (
            self.lags.get("future")[-1] if "future" in self.lags else None
        )
        return (
            min_target_lag,
            max_target_lag,
            min_past_cov_lag,
            max_past_cov_lag,
            min_future_cov_lag,
            max_future_cov_lag,
        )

    @property
    def min_train_series_length(self) -> int:
        return max(
            3,
            -self.lags["target"][0] + self.output_chunk_length
            if "target" in self.lags
            else self.output_chunk_length,
        )

    @property
    def min_train_samples(self) -> int:
        return 2

    def get_multioutput_estimator(self, horizon, target_dim):
        raise_if_not(
            isinstance(self.model, MultiOutputRegressor),
            "The sklearn model is not a MultiOutputRegressor object.",
        )

        return self.model.estimators_[horizon + target_dim]

    def _get_last_prediction_time(self, series, forecast_horizon, overlap_end):
        # overrides the ForecastingModel _get_last_prediction_time, taking care of future lags if any
        extra_shift = max(0, max(lags[-1] for lags in self.lags.values()))

        if overlap_end:
            last_valid_pred_time = series.time_index[-1 - extra_shift]
        else:
            last_valid_pred_time = series.time_index[-forecast_horizon - extra_shift]

        return last_valid_pred_time

    def _create_lagged_data(
        self, target_series, past_covariates, future_covariates, max_samples_per_ts
    ):
        lags = self.lags.get("target")
        lags_past_covariates = self.lags.get("past")
        lags_future_covariates = self.lags.get("future")

        features, labels, _ = create_lagged_training_data(
            target_series=target_series,
            output_chunk_length=self.output_chunk_length,
            past_covariates=past_covariates,
            future_covariates=future_covariates,
            lags=lags,
            lags_past_covariates=lags_past_covariates,
            lags_future_covariates=lags_future_covariates,
            max_samples_per_ts=max_samples_per_ts,
            multi_models=self.multi_models,
            check_inputs=False,
            concatenate=False,
        )

        for i, (X_i, y_i) in enumerate(zip(features, labels)):
            features[i] = X_i[:, :, 0]
            labels[i] = y_i[:, :, 0]

        features, static_covariates_shape = add_static_covariates_to_lagged_data(
            features,
            target_series,
            uses_static_covariates=self.uses_static_covariates,
            last_shape=None,
        )
        self._static_covariates_shape = static_covariates_shape

        training_samples = np.concatenate(features, axis=0)
        training_labels = np.concatenate(labels, axis=0)

        return training_samples, training_labels

    def _fit_model(
        self,
        target_series,
        past_covariates,
        future_covariates,
        max_samples_per_ts,
        **kwargs,
    ):
        """
        Function that fit the model. Deriving classes can override this method for adding additional parameters (e.g.,
        adding validation data), keeping the sanity checks on series performed by fit().
        """

        training_samples, training_labels = self._create_lagged_data(
            target_series,
            past_covariates,
            future_covariates,
            max_samples_per_ts,
        )

        # if training_labels is of shape (n_samples, 1) flatten it to shape (n_samples,)
        if len(training_labels.shape) == 2 and training_labels.shape[1] == 1:
            training_labels = training_labels.ravel()
        self.model.fit(training_samples, training_labels, **kwargs)

        # generate and store the lagged components names (for feature importance analysis)
        self._lagged_feature_names, _ = create_lagged_component_names(
            target_series=target_series,
            past_covariates=past_covariates,
            future_covariates=future_covariates,
            lags=self.lags.get("target"),
            lags_past_covariates=self.lags.get("past"),
            lags_future_covariates=self.lags.get("future"),
            output_chunk_length=self.output_chunk_length,
            concatenate=False,
            use_static_covariates=self.uses_static_covariates,
        )

    def fit(
        self,
        series: Union[TimeSeries, Sequence[TimeSeries]],
        past_covariates: Optional[Union[TimeSeries, Sequence[TimeSeries]]] = None,
        future_covariates: Optional[Union[TimeSeries, Sequence[TimeSeries]]] = None,
        max_samples_per_ts: Optional[int] = None,
        n_jobs_multioutput_wrapper: Optional[int] = None,
        **kwargs,
    ):
        """
        Fit/train the model on one or multiple series.

        Parameters
        ----------
        series
            TimeSeries or Sequence[TimeSeries] object containing the target values.
        past_covariates
            Optionally, a series or sequence of series specifying past-observed covariates
        future_covariates
            Optionally, a series or sequence of series specifying future-known covariates
        max_samples_per_ts
            This is an integer upper bound on the number of tuples that can be produced
            per time series. It can be used in order to have an upper bound on the total size of the dataset and
            ensure proper sampling. If `None`, it will read all of the individual time series in advance (at dataset
            creation) to know their sizes, which might be expensive on big datasets.
            If some series turn out to have a length that would allow more than `max_samples_per_ts`, only the
            most recent `max_samples_per_ts` samples will be considered.
        n_jobs_multioutput_wrapper
            Number of jobs of the MultiOutputRegressor wrapper to run in parallel. Only used if the model doesn't
            support multi-output regression natively.
        **kwargs
            Additional keyword arguments passed to the `fit` method of the model.
        """
        # guarantee that all inputs are either list of TimeSeries or None
        series = series2seq(series)
        past_covariates = series2seq(past_covariates)
        future_covariates = series2seq(future_covariates)

        self.encoders = self.initialize_encoders()
        if self.encoders.encoding_available:
            past_covariates, future_covariates = self.generate_fit_encodings(
                series=series,
                past_covariates=past_covariates,
                future_covariates=future_covariates,
            )

        if past_covariates is not None:
            self._uses_past_covariates = True
        if future_covariates is not None:
            self._uses_future_covariates = True
        if (
            get_single_series(series).static_covariates is not None
            and self.supports_static_covariates
            and self.considers_static_covariates
        ):
            self._uses_static_covariates = True

        for covs, name in zip([past_covariates, future_covariates], ["past", "future"]):
            raise_if(
                covs is not None and name not in self.lags,
                f"`{name}_covariates` not None in `fit()` method call, but `lags_{name}_covariates` is None in "
                f"constructor.",
            )

            raise_if(
                covs is None and name in self.lags,
                f"`{name}_covariates` is None in `fit()` method call, but `lags_{name}_covariates` is not None in "
                "constructor.",
            )

        # saving the dims of all input series to check at prediction time
        self.input_dim = {
            "target": series[0].width,
            "past": past_covariates[0].width if past_covariates else None,
            "future": future_covariates[0].width if future_covariates else None,
        }

        # if multi-output regression
        if not series[0].is_univariate or (
            self.output_chunk_length > 1 and self.multi_models
        ):
            # and model isn't wrapped already
            if not isinstance(self.model, MultiOutputRegressor):
                # check whether model supports multi-output regression natively
                if not (
                    callable(getattr(self.model, "_get_tags", None))
                    and isinstance(self.model._get_tags(), dict)
                    and self.model._get_tags().get("multioutput")
                ):
                    # if not, wrap model with MultiOutputRegressor
                    self.model = MultiOutputRegressor(
                        self.model, n_jobs=n_jobs_multioutput_wrapper
                    )
                elif isinstance(self.model, CatBoostRegressor):
                    if (
                        self.model.get_params()["loss_function"]
                        == "RMSEWithUncertainty"
                    ):
                        self.model = MultiOutputRegressor(
                            self.model, n_jobs=n_jobs_multioutput_wrapper
                        )

        # warn if n_jobs_multioutput_wrapper was provided but not used
        if (
            not isinstance(self.model, MultiOutputRegressor)
            and n_jobs_multioutput_wrapper is not None
        ):
            logger.warning("Provided `n_jobs_multioutput_wrapper` wasn't used.")

        super().fit(
            series=seq2series(series),
            past_covariates=seq2series(past_covariates),
            future_covariates=seq2series(future_covariates),
        )

        self._fit_model(
            series, past_covariates, future_covariates, max_samples_per_ts, **kwargs
        )

        return self

    def predict(
        self,
        n: int,
        series: Optional[Union[TimeSeries, Sequence[TimeSeries]]] = None,
        past_covariates: Optional[Union[TimeSeries, Sequence[TimeSeries]]] = None,
        future_covariates: Optional[Union[TimeSeries, Sequence[TimeSeries]]] = None,
        num_samples: int = 1,
        verbose: bool = False,
        **kwargs,
    ) -> Union[TimeSeries, Sequence[TimeSeries]]:
        """Forecasts values for `n` time steps after the end of the series.

        Parameters
        ----------
        n : int
            Forecast horizon - the number of time steps after the end of the series for which to produce predictions.
        series : TimeSeries or list of TimeSeries, optional
            Optionally, one or several input `TimeSeries`, representing the history of the target series whose future
            is to be predicted. If specified, the method returns the forecasts of these series. Otherwise, the method
            returns the forecast of the (single) training series.
        past_covariates : TimeSeries or list of TimeSeries, optional
            Optionally, the past-observed covariates series needed as inputs for the model.
            They must match the covariates used for training in terms of dimension and type.
        future_covariates : TimeSeries or list of TimeSeries, optional
            Optionally, the future-known covariates series needed as inputs for the model.
            They must match the covariates used for training in terms of dimension and type.
        num_samples : int, default: 1
            Currently this parameter is ignored for regression models.
        **kwargs : dict, optional
            Additional keyword arguments passed to the `predict` method of the model. Only works with
            univariate target series.
        """
        raise_if(
            not self._is_probabilistic() and num_samples > 1,
            "`num_samples > 1` is only supported for probabilistic models.",
            logger,
        )

        if series is None:
            # then there must be a single TS, and that was saved in super().fit as self.training_series
            raise_if(
                self.training_series is None,
                "Input series has to be provided after fitting on multiple series.",
            )
            series = self.training_series

        called_with_single_series = True if isinstance(series, TimeSeries) else False

        # guarantee that all inputs are either list of TimeSeries or None
        series = series2seq(series)
        past_covariates = series2seq(past_covariates)
        future_covariates = series2seq(future_covariates)

        if self.encoders.encoding_available:
            past_covariates, future_covariates = self.generate_predict_encodings(
                n=n,
                series=series,
                past_covariates=past_covariates,
                future_covariates=future_covariates,
            )

        if past_covariates is None and self.past_covariate_series is not None:
            past_covariates = series2seq(self.past_covariate_series)
        if future_covariates is None and self.future_covariate_series is not None:
            future_covariates = series2seq(self.future_covariate_series)

        super().predict(n, series, past_covariates, future_covariates, num_samples)

        # check that the input sizes of the target series and covariates match
        pred_input_dim = {
            "target": series[0].width,
            "past": past_covariates[0].width if past_covariates else None,
            "future": future_covariates[0].width if future_covariates else None,
        }
        raise_if_not(
            pred_input_dim == self.input_dim,
            f"The number of components of the target series and the covariates provided for prediction doesn't "
            f"match the number of components of the target series and the covariates this model has been "
            f"trained on.\n"
            f"Provided number of components for prediction: {pred_input_dim}\n"
            f"Provided number of components for training: {self.input_dim}",
        )

        # prediction preprocessing
        covariates = {
            "past": (past_covariates, self.lags.get("past")),
            "future": (future_covariates, self.lags.get("future")),
        }

        # prepare one_shot shift and step
        if self.multi_models:
            shift = 0
            step = self.output_chunk_length
        else:
            shift = self.output_chunk_length - 1
            step = 1

        # dictionary containing covariate data over time span required for prediction
        covariate_matrices = {}
        # dictionary containing covariate lags relative to minimum covariate lag
        relative_cov_lags = {}
        for cov_type, (covs, lags) in covariates.items():
            if covs is None:
                continue

            relative_cov_lags[cov_type] = np.array(lags) - lags[0]
            covariate_matrices[cov_type] = []
            for idx, (ts, cov) in enumerate(zip(series, covs)):
                # how many steps to go back from end of target series for start of covariates
                steps_back = -(min(lags) + 1) + shift
                lags_diff = max(lags) - min(lags) + 1
                # over how many steps the covariates range
                n_steps = lags_diff + max(0, n - self.output_chunk_length) + shift

                # calculate first and last required covariate time steps
                start_ts = ts.end_time() - ts.freq * steps_back
                end_ts = start_ts + ts.freq * (n_steps - 1)

                # check for sufficient covariate data
                if not (cov.start_time() <= start_ts and cov.end_time() >= end_ts):
                    raise_log(
                        ValueError(
                            f"The corresponding {cov_type}_covariate of the series at index {idx} isn't sufficiently "
                            f"long. Given horizon `n={n}`, `min(lags_{cov_type}_covariates)={lags[0]}`, "
                            f"`max(lags_{cov_type}_covariates)={lags[-1]}` and "
                            f"`output_chunk_length={self.output_chunk_length}`, the {cov_type}_covariate has to range "
                            f"from {start_ts} until {end_ts} (inclusive), but it ranges only from {cov.start_time()} "
                            f"until {cov.end_time()}."
                        ),
                        logger=logger,
                    )

                # use slice() instead of [] as for integer-indexed series [] does not act on time index
                # for range indexes, we make the end timestamp inclusive here
                end_ts = end_ts + ts.freq if ts.has_range_index else end_ts
                covariate_matrices[cov_type].append(
                    cov.slice(start_ts, end_ts).values(copy=False)
                )

            covariate_matrices[cov_type] = np.stack(covariate_matrices[cov_type])

        series_matrix = None
        if "target" in self.lags:
            series_matrix = np.stack(
                [
                    ts.values(copy=False)[self.lags["target"][0] - shift :, :]
                    for ts in series
                ]
            )

        # repeat series_matrix to shape (num_samples * num_series, n_lags, n_components)
        # [series 0 sample 0, series 0 sample 1, ..., series n sample k]
        series_matrix = np.repeat(series_matrix, num_samples, axis=0)

        # same for covariate matrices
        for cov_type, data in covariate_matrices.items():
            covariate_matrices[cov_type] = np.repeat(data, num_samples, axis=0)
        # prediction
        predictions = []
        last_step_shift = 0

        # t_pred indicates the number of time steps after the first prediction
        for t_pred in range(0, n, step):
            # in case of autoregressive forecast `(t_pred > 0)` and if `n` is not a round multiple of `step`,
            # we have to step back `step` from `n` in the last iteration
            if 0 < n - t_pred < step and t_pred > 0:
                last_step_shift = t_pred - (n - step)
                t_pred = n - step

            np_X = []
            # retrieve target lags
            if "target" in self.lags:
                if predictions:
                    series_matrix = np.concatenate(
                        [series_matrix, predictions[-1]], axis=1
                    )
                np_X.append(
                    series_matrix[
                        :,
                        [
                            lag - (shift + last_step_shift)
                            for lag in self.lags["target"]
                        ],
                    ].reshape(len(series) * num_samples, -1)
                )
            # retrieve covariate lags, enforce order (dict only preserves insertion order for python 3.6+)
            for cov_type in ["past", "future"]:
                if cov_type in covariate_matrices:
                    np_X.append(
                        covariate_matrices[cov_type][
                            :, relative_cov_lags[cov_type] + t_pred
                        ].reshape(len(series) * num_samples, -1)
                    )

            # concatenate retrieved lags
            X = np.concatenate(np_X, axis=1)
            # Need to split up `X` into three equally-sized sub-blocks
            # corresponding to each timeseries in `series`, so that
            # static covariates can be added to each block; valid since
            # each block contains same number of observations:
            X_blocks = np.split(X, len(series), axis=0)
            X_blocks, _ = add_static_covariates_to_lagged_data(
                X_blocks,
                series,
                uses_static_covariates=self.uses_static_covariates,
                last_shape=self._static_covariates_shape,
            )
            X = np.concatenate(X_blocks, axis=0)

            # X has shape (n_series * n_samples, n_regression_features)
            prediction = self._predict_and_sample(X, num_samples, **kwargs)
            # prediction shape (n_series * n_samples, output_chunk_length, n_components)
            # append prediction to final predictions
            predictions.append(prediction[:, last_step_shift:])

        # concatenate and use first n points as prediction
        predictions = np.concatenate(predictions, axis=1)[:, :n]

        # bring into correct shape: (n_series, output_chunk_length, n_components, n_samples)
        predictions = np.moveaxis(
            predictions.reshape(len(series), num_samples, n, -1), 1, -1
        )
        # build time series from the predicted values starting after end of series
        predictions = [
            self._build_forecast_series(row, input_tgt)
            for row, input_tgt in zip(predictions, series)
        ]

        return predictions[0] if called_with_single_series else predictions

    def _predict_and_sample(
        self, x: np.ndarray, num_samples: int, **kwargs
    ) -> np.ndarray:
        prediction = self.model.predict(x, **kwargs)
        k = x.shape[0]

        return prediction.reshape(k, self.pred_dim, -1)

    @property
    def lagged_feature_names(self) -> Optional[List[str]]:
        """The lagged feature names the model has been trained on.

        The naming convention for target, past and future covariates is: ``"{name}_{type}_lag{i}"``, where:

            - ``{name}`` the component name of the (first) series
            - ``{type}`` is the feature type, one of "target", "pastcov", and "futcov"
            - ``{i}`` is the lag value

        The naming convention for static covariates is: ``"{name}_statcov_target_{comp}"``, where:

            - ``{name}`` the static covariate name of the (first) series
            - ``{comp}`` the target component name of the (first) that the static covariate act on. If the static
                covariate acts globally on a multivariate target series, will show "global".
        """
        return self._lagged_feature_names

    def __str__(self):
        return self.model.__str__()

    @property
    def supports_static_covariates(self) -> bool:
        return True


class _LikelihoodMixin:
    """
    A class containing functions supporting quantile, poisson and gaussian regression, to be used as a mixin for some
    `RegressionModel` subclasses.
    """

    @staticmethod
    def _check_likelihood(likelihood, available_likelihoods):
        raise_if_not(
            likelihood in available_likelihoods,
            f"If likelihood is specified it must be one of {available_likelihoods}",
        )

    @staticmethod
    def _get_model_container():
        return _QuantileModelContainer()

    @staticmethod
    def _prepare_quantiles(quantiles):
        if quantiles is None:
            quantiles = [
                0.01,
                0.05,
                0.1,
                0.25,
                0.5,
                0.75,
                0.9,
                0.95,
                0.99,
            ]
        else:
            quantiles = sorted(quantiles)
            _check_quantiles(quantiles)
        median_idx = quantiles.index(0.5)

        return quantiles, median_idx

    def _predict_quantiles(
        self, x: np.ndarray, num_samples: int, **kwargs
    ) -> np.ndarray:
        """
        X is of shape (n_series * n_samples, n_regression_features)
        """
        k = x.shape[0]

        if num_samples == 1:
            # return median
            fitted = self._model_container[0.5]
            return fitted.predict(x, **kwargs).reshape(k, self.pred_dim, -1)

        model_outputs = []
        for quantile, fitted in self._model_container.items():
            self.model = fitted
            # model output has shape (n_series * n_samples, output_chunk_length, n_components)
            model_output = fitted.predict(x, **kwargs).reshape(k, self.pred_dim, -1)
            model_outputs.append(model_output)
        model_outputs = np.stack(model_outputs, axis=-1)
        # model_outputs has shape (n_series * n_samples, output_chunk_length, n_components, n_quantiles)

        sampled = self._quantile_sampling(model_outputs)

        # sampled has shape (n_series * n_samples, output_chunk_length, n_components)

        return sampled

    def _predict_normal(self, x: np.ndarray, num_samples: int, **kwargs) -> np.ndarray:
        """Method intended for CatBoost's RMSEWithUncertainty loss. Returns samples
        computed from double-valued inputs [mean, variance].
        X is of shape (n_series * n_samples, n_regression_features)
        """
        k = x.shape[0]

        # model_output shape:
        # if univariate & output_chunk_length = 1: (num_samples, 2)
        # else: (2, num_samples, n_components * output_chunk_length)
        # where the axis with 2 dims is mu, sigma
        model_output = self.model.predict(x, **kwargs)
        output_dim = len(model_output.shape)

        # deterministic case: we return the mean only
        if num_samples == 1:
            # univariate & single-chunk output
            if output_dim <= 2:
                output_slice = model_output[:, 0]
            else:
                output_slice = model_output[0, :, :]

            return output_slice.reshape(k, self.pred_dim, -1)

        # probabilistic case
        # univariate & single-chunk output
        if output_dim <= 2:
            # embedding well shaped 2D output into 3D
            model_output = np.expand_dims(model_output, axis=0)

        else:
            # we transpose to get mu, sigma couples on last axis
            # shape becomes: (n_components * output_chunk_length, num_samples, 2)
            model_output = model_output.transpose()

        return self._normal_sampling(model_output, num_samples)

    def _normal_sampling(self, model_output: np.ndarray, n_samples: int) -> np.ndarray:
        """Sampling method for CatBoost's [mean, variance] output.
        model_output is of shape (n_components * output_chunk_length, n_samples, 2),
        where the last 2 dimensions are mu and sigma.
        """
        shape = model_output.shape
        chunk_len = self.pred_dim

        # treating each component separately
        mu_sigma_list = [model_output[i, :, :] for i in range(shape[0])]

        list_of_samples = [
            self._rng.normal(
                mu_sigma[:, 0],  # mean vector
                mu_sigma[:, 1],  # diagonal covariance matrix
            )
            for mu_sigma in mu_sigma_list
        ]

        samples_transposed = np.array(list_of_samples).transpose()
        samples_reshaped = samples_transposed.reshape(n_samples, chunk_len, -1)

        return samples_reshaped

    def _predict_poisson(self, x: np.ndarray, num_samples: int, **kwargs) -> np.ndarray:
        """
        X is of shape (n_series * n_samples, n_regression_features)
        """
        k = x.shape[0]

        model_output = self.model.predict(x, **kwargs).reshape(k, self.pred_dim, -1)
        if num_samples == 1:
            return model_output

        return self._poisson_sampling(model_output)

    def _poisson_sampling(self, model_output: np.ndarray) -> np.ndarray:
        """
        Model_output is of shape (n_series * n_samples, output_chunk_length, n_components)
        """

        return self._rng.poisson(lam=model_output).astype(float)

    def _quantile_sampling(self, model_output: np.ndarray) -> np.ndarray:
        """
        Sample uniformly between [0, 1] (for each batch example) and return the linear interpolation between the fitted
        quantiles closest to the sampled value.

        model_output is of shape (batch_size, n_timesteps, n_components, n_quantiles)
        """
        num_samples, n_timesteps, n_components, n_quantiles = model_output.shape

        # obtain samples
        probs = self._rng.uniform(
            size=(
                num_samples,
                n_timesteps,
                n_components,
                1,
            )
        )

        # add dummy dim
        probas = np.expand_dims(probs, axis=-2)

        # tile and transpose
        p = np.tile(probas, (1, 1, 1, n_quantiles, 1)).transpose((0, 1, 2, 4, 3))

        # prepare quantiles
        tquantiles = np.array(self.quantiles).reshape((1, 1, 1, -1))

        # calculate index of the largest quantile smaller than the sampled value
        left_idx = np.sum(p > tquantiles, axis=-1)

        # obtain index of the smallest quantile larger than the sampled value
        right_idx = left_idx + 1

        # repeat the model output on the edges
        repeat_count = [1] * n_quantiles
        repeat_count[0] = 2
        repeat_count[-1] = 2
        repeat_count = np.array(repeat_count)
        shifted_output = np.repeat(model_output, repeat_count, axis=-1)

        # obtain model output values corresponding to the quantiles left and right of the sampled value
        left_value = np.take_along_axis(shifted_output, left_idx, axis=-1)
        right_value = np.take_along_axis(shifted_output, right_idx, axis=-1)

        # add 0 and 1 to quantiles
        ext_quantiles = [0.0] + self.quantiles + [1.0]
        expanded_q = np.tile(np.array(ext_quantiles), left_idx.shape)

        # calculate closest quantiles to the sampled value
        left_q = np.take_along_axis(expanded_q, left_idx, axis=-1)
        right_q = np.take_along_axis(expanded_q, right_idx, axis=-1)

        # linear interpolation
        weights = (probs - left_q) / (right_q - left_q)
        inter = left_value + weights * (right_value - left_value)

        return inter.squeeze(-1)


class _QuantileModelContainer(OrderedDict):
    def __init__(self):
        super().__init__()


class RegressionModelWithCategoricalCovariates(RegressionModel):
    def __init__(
        self,
        lags: Union[int, list] = None,
        lags_past_covariates: Union[int, List[int]] = None,
        lags_future_covariates: Union[Tuple[int, int], List[int]] = None,
        output_chunk_length: int = 1,
        add_encoders: Optional[dict] = None,
        model=None,
        multi_models: Optional[bool] = True,
        use_static_covariates: bool = True,
        categorical_past_covariates: Optional[Union[str, List[str]]] = None,
        categorical_future_covariates: Optional[Union[str, List[str]]] = None,
        categorical_static_covariates: Optional[Union[str, List[str]]] = None,
    ):
        """
        Extension of `RegressionModel` for regression models that support categorical covariates.

        Parameters
        ----------
        lags
            Lagged target values used to predict the next time step. If an integer is given the last `lags` past lags
            are used (from -1 backward). Otherwise, a list of integers with lags is required (each lag must be < 0).
        lags_past_covariates
            Number of lagged past_covariates values used to predict the next time step. If an integer is given the last
            `lags_past_covariates` past lags are used (inclusive, starting from lag -1). Otherwise a list of integers
            with lags < 0 is required.
        lags_future_covariates
            Number of lagged future_covariates values used to predict the next time step. If a tuple (past, future) is
            given the last `past` lags in the past are used (inclusive, starting from lag -1) along with the first
            `future` future lags (starting from 0 - the prediction time - up to `future - 1` included). Otherwise a list
            of integers with lags is required.
        output_chunk_length
            Number of time steps predicted at once by the internal regression model. Does not have to equal the forecast
            horizon `n` used in `predict()`. However, setting `output_chunk_length` equal to the forecast horizon may
            be useful if the covariates don't extend far enough into the future.
        add_encoders
            A large number of past and future covariates can be automatically generated with `add_encoders`.
            This can be done by adding multiple pre-defined index encoders and/or custom user-made functions that
            will be used as index encoders. Additionally, a transformer such as Darts' :class:`Scaler` can be added to
            transform the generated covariates. This happens all under one hood and only needs to be specified at
            model creation.
            Read :meth:`SequentialEncoder <darts.dataprocessing.encoders.SequentialEncoder>` to find out more about
            ``add_encoders``. Default: ``None``. An example showing some of ``add_encoders`` features:

            .. highlight:: python
            .. code-block:: python

                add_encoders={
                    'cyclic': {'future': ['month']},
                    'datetime_attribute': {'future': ['hour', 'dayofweek']},
                    'position': {'past': ['relative'], 'future': ['relative']},
                    'custom': {'past': [lambda idx: (idx.year - 1950) / 50]},
                    'transformer': Scaler()
                }
            ..
        model
            Scikit-learn-like model with ``fit()`` and ``predict()`` methods. Also possible to use model that doesn't
            support multi-output regression for multivariate timeseries, in which case one regressor
            will be used per component in the multivariate series.
            If None, defaults to: ``sklearn.linear_model.LinearRegression(n_jobs=-1)``.
        multi_models
            If True, a separate model will be trained for each future lag to predict. If False, a single model is
            trained to predict at step 'output_chunk_length' in the future. Default: True.
        use_static_covariates
            Whether the model should use static covariate information in case the input `series` passed to ``fit()``
            contain static covariates. If ``True``, and static covariates are available at fitting time, will enforce
            that all target `series` have the same static covariate dimensionality in ``fit()`` and ``predict()`.
        categorical_past_covariates
            Optionally, component name or list of component names specifying the past covariates that should be treated
            as categorical.
        categorical_future_covariates
            Optionally, component name or list of component names specifying the future covariates that should be
            treated as categorical.
        categorical_static_covariates
            Optionally, string or list of strings specifying the static covariates that should be treated as
            categorical.
        """
        super().__init__(
            lags=lags,
            lags_past_covariates=lags_past_covariates,
            lags_future_covariates=lags_future_covariates,
            output_chunk_length=output_chunk_length,
            add_encoders=add_encoders,
            model=model,
            multi_models=multi_models,
            use_static_covariates=use_static_covariates,
        )
        self.categorical_past_covariates = (
            [categorical_past_covariates]
            if isinstance(categorical_past_covariates, str)
            else categorical_past_covariates
        )
        self.categorical_future_covariates = (
            [categorical_future_covariates]
            if isinstance(categorical_future_covariates, str)
            else categorical_future_covariates
        )
        self.categorical_static_covariates = (
            [categorical_static_covariates]
            if isinstance(categorical_static_covariates, str)
            else categorical_static_covariates
        )

    def fit(
        self,
        series: Union[TimeSeries, Sequence[TimeSeries]],
        past_covariates: Optional[Union[TimeSeries, Sequence[TimeSeries]]] = None,
        future_covariates: Optional[Union[TimeSeries, Sequence[TimeSeries]]] = None,
        max_samples_per_ts: Optional[int] = None,
        n_jobs_multioutput_wrapper: Optional[int] = None,
        **kwargs,
    ):
        self._validate_categorical_covariates(
            series=series,
            past_covariates=past_covariates,
            future_covariates=future_covariates,
        )
        super().fit(
            series=series,
            past_covariates=past_covariates,
            future_covariates=future_covariates,
            max_samples_per_ts=max_samples_per_ts,
            n_jobs_multioutput_wrapper=n_jobs_multioutput_wrapper,
            **kwargs,
        )

    @property
    def _categorical_fit_param(self) -> Tuple[str, Any]:
        """
        Returns the name, and default value of the categorical features parameter from model's `fit` method .
        Can be overridden in subclasses.
        """
        return "categorical_feature", "auto"

    def _validate_categorical_covariates(
        self,
        series: Union[TimeSeries, Sequence[TimeSeries]],
        past_covariates: Optional[Union[TimeSeries, Sequence[TimeSeries]]] = None,
        future_covariates: Optional[Union[TimeSeries, Sequence[TimeSeries]]] = None,
    ) -> None:
        """
        Checks that the categorical covariates are valid. Specifically, checks that the categorical covariates
        of the model are a subset of all covariates.

        Parameters
        ----------
        series
            TimeSeries or Sequence[TimeSeries] object containing the target values.
        past_covariates
            Optionally, a series or sequence of series specifying past-observed covariates
        future_covariates
            Optionally, a series or sequence of series specifying future-known covariates
        """
        for categorical_covariates, covariates, cov_type in zip(
            [self.categorical_past_covariates, self.categorical_future_covariates],
            [past_covariates, future_covariates],
            ["past_covariates", "future_covariates"],
        ):
            if categorical_covariates:
                if not covariates:
                    raise_log(
                        ValueError(
                            f"`categorical_{cov_type}` were declared at model creation but no "
                            f"`{cov_type}` are passed to the `fit()` call."
                        ),
                    )
                s = get_single_series(covariates)
                if not set(categorical_covariates).issubset(set(s.components)):
                    raise_log(
                        ValueError(
                            f"Some `categorical_{cov_type}` components "
                            f"({set(categorical_covariates) - set(s.components)}) "
                            f"declared at model creation are not present in the `{cov_type}` "
                            f"passed to the `fit()` call."
                        )
                    )
        if self.categorical_static_covariates:
            s = get_single_series(series)
            covariates = s.static_covariates
            if not s.has_static_covariates:
                raise_log(
                    ValueError(
                        "`categorical_static_covariates` were declared at model creation but `series`"
                        "passed to the `fit()` call does not contain `static_covariates`."
                    ),
                )
            if not set(self.categorical_static_covariates).issubset(
                set(covariates.columns)
            ):
                raise_log(
                    ValueError(
                        f"Some `categorical_static_covariates` components "
                        f"({set(self.categorical_static_covariates) - set(covariates.columns)}) "
                        f"declared at model creation are not present in the series' `static_covariates` "
                        f"passed to the `fit()` call."
                    )
                )

    def _get_categorical_features(
        self,
        series: Union[List[TimeSeries], TimeSeries],
        past_covariates: Optional[Union[List[TimeSeries], TimeSeries]] = None,
        future_covariates: Optional[Union[List[TimeSeries], TimeSeries]] = None,
    ) -> Tuple[List[int], List[str]]:
        """
        Returns the indices and column names of the categorical features in the regression model.

        Steps:
        1. Get the list of features used in the model. We keep the creation order of the different lags/features
            in create_lagged_data.
        2. Get the indices of the categorical features in the list of features.
        """

        categorical_covariates = (
            (
                self.categorical_past_covariates
                if self.categorical_past_covariates
                else []
            )
            + (
                self.categorical_future_covariates
                if self.categorical_future_covariates
                else []
            )
            + (
                self.categorical_static_covariates
                if self.categorical_static_covariates
                else []
            )
        )

        if not categorical_covariates:
            return [], []
        else:
            target_ts = get_single_series(series)
            past_covs_ts = get_single_series(past_covariates)
            fut_covs_ts = get_single_series(future_covariates)

            # We keep the creation order of the different lags/features in create_lagged_data
            feature_list = (
                [
                    f"target_{component}_lag{lag}"
                    for lag in self.lags.get("target", [])
                    for component in target_ts.components
                ]
                + [
                    f"past_cov_{component}_lag{lag}"
                    for lag in self.lags.get("past", [])
                    for component in past_covs_ts.components
                ]
                + [
                    f"fut_cov_{component}_lag{lag}"
                    for lag in self.lags.get("future", [])
                    for component in fut_covs_ts.components
                ]
                + (
                    list(target_ts.static_covariates.columns)
                    if target_ts.has_static_covariates
                    # if isinstance(target_ts.static_covariates, pd.DataFrame)
                    else []
                )
            )

            indices = [
                i
                for i, col in enumerate(feature_list)
                for cat in categorical_covariates
                if cat and cat in col
            ]
            col_names = [feature_list[i] for i in indices]

            return indices, col_names

    def _fit_model(
        self,
        target_series,
        past_covariates,
        future_covariates,
        max_samples_per_ts,
        **kwargs,
    ):
        """
        Custom fit function for `RegressionModelWithCategoricalCovariates` models, adding logic to let the model
        handle categorical features directly.
        """
        cat_col_indices, _ = self._get_categorical_features(
            target_series,
            past_covariates,
            future_covariates,
        )

        cat_param_name, cat_param_default = self._categorical_fit_param
        kwargs[cat_param_name] = (
            cat_col_indices if cat_col_indices else cat_param_default
        )
        super()._fit_model(
            target_series=target_series,
            past_covariates=past_covariates,
            future_covariates=future_covariates,
            max_samples_per_ts=max_samples_per_ts,
            **kwargs,
        )



import os

import pytest

from darts import TimeSeries
from darts.datasets import (
    _DEFAULT_PATH,
    AirPassengersDataset,
    AusBeerDataset,
    AustralianTourismDataset,
    ElectricityDataset,
    EnergyDataset,
    ETTh1Dataset,
    ETTh2Dataset,
    ETTm1Dataset,
    ETTm2Dataset,
    ExchangeRateDataset,
    GasRateCO2Dataset,
    HeartRateDataset,
    IceCreamHeaterDataset,
    ILINetDataset,
    MonthlyMilkDataset,
    MonthlyMilkIncompleteDataset,
    SunspotsDataset,
    TaylorDataset,
    TemperatureDataset,
    TrafficDataset,
    UberTLCDataset,
    USGasolineDataset,
    WeatherDataset,
    WineDataset,
    WoolyDataset,
)
from darts.datasets.dataset_loaders import (
    DatasetLoader,
    DatasetLoaderCSV,
    DatasetLoaderMetadata,
    DatasetLoadingException,
)
from darts.tests.base_test_class import DartsBaseTestClass

datasets = [
    AirPassengersDataset,
    AusBeerDataset,
    AustralianTourismDataset,
    EnergyDataset,
    HeartRateDataset,
    IceCreamHeaterDataset,
    MonthlyMilkDataset,
    SunspotsDataset,
    TaylorDataset,
    TemperatureDataset,
    USGasolineDataset,
    WineDataset,
    WoolyDataset,
    GasRateCO2Dataset,
    MonthlyMilkIncompleteDataset,
    ETTh1Dataset,
    ETTh2Dataset,
    ETTm1Dataset,
    ETTm2Dataset,
    ElectricityDataset,
    UberTLCDataset,
    ILINetDataset,
    ExchangeRateDataset,
    TrafficDataset,
    WeatherDataset,
]

_DEFAULT_PATH_TEST = _DEFAULT_PATH + "/tests"

width_datasets = [1, 1, 96, 28, 1, 2, 1, 1, 1, 1, 1, 1, 1, 2, 1, 7, 7, 7, 7, 370, 262]

wrong_hash_dataset = DatasetLoaderCSV(
    metadata=DatasetLoaderMetadata(
        "wrong_hash",
        uri="https://raw.githubusercontent.com/unit8co/darts/master/datasets/data/AirPassengers.csv",
        hash="will fail",
        header_time="Month",
        format_time="%Y-%m",
    )
)

wrong_url_dataset = DatasetLoaderCSV(
    metadata=DatasetLoaderMetadata(
        "wrong_url",
        uri="https://AirPassengers.csv",
        hash="167ffa96204a2b47339c21eea25baf32",
        header_time="Month",
        format_time="%Y-%m",
    )
)

wrong_zip_url_dataset = DatasetLoaderCSV(
    metadata=DatasetLoaderMetadata(
        "wrong_zip_url",
        uri="https://Electricity.zip",
        hash="d17748042ea98fc9c5fb4db0946d5fa4",
        header_time="Unnamed: 0",
        format_time="%Y-%m-%d %H:%M:%S",
        pre_process_zipped_csv_fn=lambda x: x,
    )
)

no_pre_process_fn_dataset = DatasetLoaderCSV(
    metadata=DatasetLoaderMetadata(
        "no_pre_process_fn",
        uri=_DEFAULT_PATH_TEST + "/test.zip",
        hash="167ffa96204a2b47339c21eea25baf32",
        header_time="Month",
        pre_process_zipped_csv_fn=None,
    )
)
ele_multi_series_dataset = DatasetLoaderCSV(
    metadata=DatasetLoaderMetadata(
        "Electricity_test.csv",
        uri=_DEFAULT_PATH_TEST + "/Electricity_test.csv",
        hash="e036be148b06dacf2bb78b4647e6ea2b",
        header_time="Time",
        pre_process_zipped_csv_fn=None,
    )
)


class DatasetLoaderTestCase(DartsBaseTestClass):
    def tearDown(self):
        # we need to remove the cached datasets between each test
        default_directory = DatasetLoader._DEFAULT_DIRECTORY
        for f in os.listdir(default_directory):
            os.remove(os.path.join(default_directory, f))
        os.rmdir(DatasetLoader._DEFAULT_DIRECTORY)

    @pytest.mark.slow
    def test_ok_dataset(self):
        for width, dataset_cls in zip(width_datasets, datasets):
            dataset = dataset_cls()
            ts: TimeSeries = dataset.load()
            self.assertEqual(ts.width, width)

    def test_hash(self):
        with self.assertRaises(DatasetLoadingException):
            wrong_hash_dataset.load()

    def test_uri(self):
        with self.assertRaises(DatasetLoadingException):
            wrong_url_dataset.load()

    def test_zip_uri(self):
        with self.assertRaises(DatasetLoadingException):
            wrong_zip_url_dataset.load()

    def test_pre_process_fn(self):
        with self.assertRaises(DatasetLoadingException):
            no_pre_process_fn_dataset.load()

    def test_multi_series_dataset(self):
        # processing _to_multi_series takes a long time. Test function with 5 cols.
        ts = ele_multi_series_dataset.load().pd_dataframe()

        ms = ElectricityDataset()._to_multi_series(ts)
        self.assertEqual(len(ms), 5)
        self.assertEqual(len(ms[0]), 105216)

        multi_series_datasets = [
            UberTLCDataset,
            ILINetDataset,
            ExchangeRateDataset,
            TrafficDataset,
            WeatherDataset,
        ]
        for dataset in multi_series_datasets:
            ms = dataset()._to_multi_series(ts)
            self.assertEqual(len(ms), 5)
            self.assertEqual(len(ms[0]), len(ts.index))



"""
Gaussian Processes
------------------
"""

from typing import Optional

import numpy as np
from sklearn.gaussian_process import GaussianProcessRegressor
from sklearn.gaussian_process.kernels import Kernel

from darts.models.filtering.filtering_model import FilteringModel
from darts.timeseries import TimeSeries


class GaussianProcessFilter(FilteringModel):
    def __init__(self, kernel: Optional[Kernel] = None, **kwargs):
        """
        This model uses the ``GaussianProcessRegressor`` of scikit-learn to fit a Gaussian Process to the
        supplied TimeSeries. This can then be used to obtain samples from the
        Gaussian Process at the times of the TimeSeries.

        It can for instance be used to fill in missing (NaN) values from a TimeSeries.

        Parameters
        ----------
        kernel : sklearn.gaussian_process.kernels.Kernel, default: None
            The kernel specifying the covariance function of the Gaussian Process. If None is passed,
            the default in scikit-learn is used. Note that the kernel hyperparameters are optimized
            during fitting unless the bounds are marked as 'fixed'.
        **kwargs
            Additional keyword arguments passed to ``sklearn.gaussian_process.GaussianProcessRegressor``.
        """
        super().__init__()
        self.model = GaussianProcessRegressor(kernel=kernel, **kwargs)

    def filter(self, series: TimeSeries, num_samples: int = 1) -> TimeSeries:
        """
        Fits the Gaussian Process on the observations and returns samples from the Gaussian Process,
        or its mean values if `num_samples` is set to 1.

        Parameters
        ----------
        series
            The series of observations used to infer the values according to the specified Gaussian Process.
            This must be a deterministic series (containing one sample).
        num_samples: int, default: 1
            Number of times a prediction is sampled from the Gaussian Process. If set to 1,
            the mean values will be returned instead.

        Returns
        -------
        TimeSeries
            A stochastic ``TimeSeries`` sampled from the Gaussian Process, or its mean
            if `num_samples` is set to 1.
        """
        super().filter(series)

        values = series.values(copy=False)
        if series.has_datetime_index:
            times = np.arange(series.n_timesteps).reshape(-1, 1)
        else:
            times = series.time_index.values.reshape(-1, 1)

        not_nan_mask = np.all(~np.isnan(values), axis=1)

        self.model.fit(times[not_nan_mask, :], values[not_nan_mask, :])

        if num_samples == 1:
            filtered_values = self.model.predict(times)
        else:
            filtered_values = self.model.sample_y(times, n_samples=num_samples)

        filtered_values = filtered_values.reshape(len(times), -1, num_samples)

        return series.with_values(filtered_values)



"""
Utils for Anomaly Detection
---------------------------

Common functions used by anomaly_model.py, scorers.py, aggregators.py and detectors.py
"""

# TODO:
#     - change structure of eval_accuracy_from_scores and eval_accuracy_from_binary_prediction (a lot of repeated code)
#     - migrate metrics function to darts.metric
#     - check error message
#     - create a zoom option on anomalies for a show function
#     - add an option visualize: "by window", "unique", "together"
#     - create a normalize option in plot function (norm every anomaly score btw 1 and 0) -> to be seen on the same plot

from typing import Sequence, Tuple, Union

import matplotlib.pyplot as plt
import numpy as np
from sklearn.metrics import (
    accuracy_score,
    average_precision_score,
    f1_score,
    precision_score,
    recall_score,
    roc_auc_score,
)

from darts import TimeSeries
from darts.logging import get_logger, raise_if, raise_if_not

logger = get_logger(__name__)


def _assert_binary(series: TimeSeries, name_series: str):
    """Checks if series is a binary timeseries (1 and 0)"

    Parameters
    ----------
    series
        series to check for.
    name_series
        name str of the series.
    """

    raise_if_not(
        np.array_equal(
            series.values(copy=False),
            series.values(copy=False).astype(bool),
        ),
        f"Input series {name_series} must be a binary time series.",
    )


def eval_accuracy_from_scores(
    actual_anomalies: Union[TimeSeries, Sequence[TimeSeries]],
    anomaly_score: Union[TimeSeries, Sequence[TimeSeries]],
    window: Union[int, Sequence[int]] = 1,
    metric: str = "AUC_ROC",
) -> Union[float, Sequence[float], Sequence[Sequence[float]]]:
    """Scores the results against true anomalies.

    `actual_anomalies` and `anomaly_score` must have the same shape.
    `actual_anomalies` must be binary and have values belonging to the two classes (0 and 1).

    If one series is given for `actual_anomalies` and `anomaly_score` contains more than
    one series, the function will consider `actual_anomalies` as the ground truth anomalies for
    all scores in `anomaly_score`.

    Parameters
    ----------
    actual_anomalies
        The ground truth of the anomalies (1 if it is an anomaly and 0 if not).
    anomaly_score
        Series indicating how anomoulous each window of size w is.
    window
        Integer value indicating the number of past samples each point represents
        in the anomaly_score. The parameter will be used by the function
        ``_window_adjustment_anomalies()`` to transform actual_anomalies.
        If a list is given. the length must match the number of series in anomaly_score
        and actual_anomalies. If only one window is given, the value will be used for every
        series in anomaly_score and actual_anomalies.
    metric
        Optionally, Scoring function to use. Must be one of "AUC_ROC" and "AUC_PR".
        Default: "AUC_ROC"

    Returns
    -------
    Union[float, Sequence[float], Sequence[Sequence[float]]]
        Score of the anomalies score prediction
            * ``float`` if `anomaly_score` is a univariate series (dimension=1).
            * ``Sequence[float]``

                * if `anomaly_score` is a multivariate series (dimension>1),
                  returns one value per dimension.
                * if `anomaly_score` is a sequence of univariate series, returns one
                  value per series
            * ``Sequence[Sequence[float]]`` if `anomaly_score` is a sequence of
              multivariate series. Outer Sequence is over the sequence input, and the inner
              Sequence is over the dimensions of each element in the sequence input.
    """

    raise_if_not(
        metric in {"AUC_ROC", "AUC_PR"},
        "Argument `metric` must be one of 'AUC_ROC', 'AUC_PR'",
    )
    metric_fn = roc_auc_score if metric == "AUC_ROC" else average_precision_score

    list_actual_anomalies, list_anomaly_scores, list_window = (
        _to_list(actual_anomalies),
        _to_list(anomaly_score),
        _to_list(window),
    )

    if len(list_actual_anomalies) == 1 and len(list_anomaly_scores) > 1:
        list_actual_anomalies = list_actual_anomalies * len(list_anomaly_scores)

    _assert_same_length(list_actual_anomalies, list_anomaly_scores)

    if len(list_window) == 1:
        list_window = list_window * len(actual_anomalies)
    else:
        raise_if_not(
            len(list_window) == len(list_actual_anomalies),
            "The list of windows must be the same length as the list of `anomaly_score` and"
            + " `actual_anomalies`. There must be one window value for each series."
            + f" Found length {len(list_window)}, expected {len(list_actual_anomalies)}.",
        )

    sol = []
    for idx, (s_anomalies, s_score) in enumerate(
        zip(list_actual_anomalies, list_anomaly_scores)
    ):

        _assert_binary(s_anomalies, "actual_anomalies")

        sol.append(
            _eval_accuracy_from_data(
                s_anomalies, s_score, list_window[idx], metric_fn, metric
            )
        )

    if len(sol) == 1 and not isinstance(anomaly_score, Sequence):
        return sol[0]
    else:
        return sol


def eval_accuracy_from_binary_prediction(
    actual_anomalies: Union[TimeSeries, Sequence[TimeSeries]],
    binary_pred_anomalies: Union[TimeSeries, Sequence[TimeSeries]],
    window: Union[int, Sequence[int]] = 1,
    metric: str = "recall",
) -> Union[float, Sequence[float], Sequence[Sequence[float]]]:
    """Score the results against true anomalies.

    checks that `pred_anomalies` and `actual_anomalies` are the same:
        - type,
        - length,
        - number of components
        - binary and has values belonging to the two classes (1 and 0)

    If one series is given for `actual_anomalies` and `pred_anomalies` contains more than
    one series, the function will consider `actual_anomalies` as the true anomalies for
    all scores in `anomaly_score`.

    Parameters
    ----------
    actual_anomalies
        The (sequence of) ground truth of the anomalies (1 if it is an anomaly and 0 if not)
    binary_pred_anomalies
        Anomaly predictions.
    window
        Integer value indicating the number of past samples each point represents
        in the pred_anomalies. The parameter will be used to transform actual_anomalies.
        If a list is given. the length must match the number of series in pred_anomalies
        and actual_anomalies. If only one window is given, the value will be used for every
        series in pred_anomalies and actual_anomalies.
    metric
        Optionally, Scoring function to use. Must be one of "recall", "precision",
        "f1", and "accuracy".
        Default: "recall"

    Returns
    -------
    Union[float, Sequence[float], Sequence[Sequence[float]]]
        Score of the anomalies prediction

            * ``float`` if `binary_pred_anomalies` is a univariate series (dimension=1).
            * ``Sequence[float]``

                * if `binary_pred_anomalies` is a multivariate series (dimension>1),
                  returns one value per dimension.
                * if `binary_pred_anomalies` is a sequence of univariate series, returns one
                  value per series
            * ``Sequence[Sequence[float]]`` if `binary_pred_anomalies` is a sequence of
              multivariate series. Outer Sequence is over the sequence input, and the inner
              Sequence is over the dimensions of each element in the sequence input.
    """

    raise_if_not(
        metric in {"recall", "precision", "f1", "accuracy"},
        "Argument `metric` must be one of 'recall', 'precision', "
        "'f1' and 'accuracy'.",
    )

    if metric == "recall":
        metric_fn = recall_score
    elif metric == "precision":
        metric_fn = precision_score
    elif metric == "f1":
        metric_fn = f1_score
    else:
        metric_fn = accuracy_score

    list_actual_anomalies, list_binary_pred_anomalies, list_window = (
        _to_list(actual_anomalies),
        _to_list(binary_pred_anomalies),
        _to_list(window),
    )

    if len(list_actual_anomalies) == 1 and len(list_binary_pred_anomalies) > 1:
        list_actual_anomalies = list_actual_anomalies * len(list_binary_pred_anomalies)

    _assert_same_length(list_actual_anomalies, list_binary_pred_anomalies)

    if len(list_window) == 1:
        list_window = list_window * len(actual_anomalies)
    else:
        raise_if_not(
            len(list_window) == len(list_actual_anomalies),
            "The list of windows must be the same length as the list of `pred_anomalies` and"
            + " `actual_anomalies`. There must be one window value for each series."
            + f" Found length {len(list_window)}, expected {len(list_actual_anomalies)}.",
        )

    sol = []
    for idx, (s_anomalies, s_pred) in enumerate(
        zip(list_actual_anomalies, list_binary_pred_anomalies)
    ):

        _assert_binary(s_pred, "pred_anomalies")
        _assert_binary(s_anomalies, "actual_anomalies")

        sol.append(
            _eval_accuracy_from_data(
                s_anomalies, s_pred, list_window[idx], metric_fn, metric
            )
        )

    if len(sol) == 1 and not isinstance(binary_pred_anomalies, Sequence):
        return sol[0]
    else:
        return sol


def _eval_accuracy_from_data(
    s_anomalies: TimeSeries,
    s_data: TimeSeries,
    window: int,
    metric_fn,
    metric_name: str,
) -> Union[float, Sequence[float]]:
    """Internal function for:
        - ``eval_accuracy_from_binary_prediction()``
        - ``eval_accuracy_from_scores()``

    Score the results against true anomalies.

    Parameters
    ----------
    actual_anomalies
        The ground truth of the anomalies (1 if it is an anomaly and 0 if not)
    s_data
        series prediction
    window
        Integer value indicating the number of past samples each point represents
        in the anomaly_score. The parameter will be used by the function
        ``_window_adjustment_anomalies()`` to transform s_anomalies.
    metric_fn
        Function to use. Can be "average_precision_score", "roc_auc_score", "accuracy_score",
        "f1_score", "precision_score" and "recall_score".
    metric_name
        Name str of the function to use. Can be "AUC_PR", "AUC_ROC", "accuracy",
        "f1", "precision" and "recall".

    Returns
    -------
    Union[float, Sequence[float]]
        Score of the anomalies prediction
            - float -> if `s_data` is a univariate series (dimension=1).
            - Sequence[float] -> if `s_data` is a multivariate series (dimension>1),
            returns one value per dimension.
    """

    _assert_timeseries(s_data, "Prediction series input")
    _assert_timeseries(s_anomalies, "actual_anomalies input")

    # if window > 1, the anomalies will be adjusted so that it can be compared timewise with s_data
    s_anomalies = _max_pooling(s_anomalies, window)

    _sanity_check_two_series(s_data, s_anomalies)

    s_data, s_anomalies = _intersect(s_data, s_anomalies)

    if metric_name == "AUC_ROC" or metric_name == "AUC_PR":

        nr_anomalies_per_component = (
            s_anomalies.sum(axis=0).values(copy=False).flatten()
        )

        raise_if(
            nr_anomalies_per_component.min() == 0,
            f"`actual_anomalies` does not contain anomalies. {metric_name} cannot be computed.",
        )

        raise_if(
            nr_anomalies_per_component.max() == len(s_anomalies),
            f"`actual_anomalies` only contains anomalies. {metric_name} cannot be computed."
            + ["", f" Consider decreasing the window size (window={window})"][
                window > 1
            ],
        )

    # TODO: could we vectorize this?
    metrics = []
    for component_idx in range(s_data.width):
        metrics.append(
            metric_fn(
                s_anomalies.all_values(copy=False)[:, component_idx],
                s_data.all_values(copy=False)[:, component_idx],
            )
        )

    if len(metrics) == 1:
        return metrics[0]
    else:
        return metrics


def _intersect(
    series_1: TimeSeries,
    series_2: TimeSeries,
) -> Tuple[TimeSeries, TimeSeries]:
    """Returns the sub-series of series_1 and of series_2 that share the same time index.
    (Intersection in time of the two time series)

    Parameters
    ----------
    series_1
        1st time series
    series_2:
        2nd time series

    Returns
    -------
    Tuple[TimeSeries, TimeSeries]
    """

    new_series_1 = series_1.slice_intersect(series_2)
    raise_if(
        len(new_series_1) == 0,
        "Time intersection between the two series must be non empty.",
    )

    return new_series_1, series_2.slice_intersect(series_1)


def _assert_timeseries(series: TimeSeries, message: str = None):
    """Checks if given input is of type Darts TimeSeries"""

    raise_if_not(
        isinstance(series, TimeSeries),
        "{} must be type darts.timeseries.TimeSeries and not {}.".format(
            message if message is not None else "Series input", type(series)
        ),
    )


def _sanity_check_two_series(
    series_1: TimeSeries,
    series_2: TimeSeries,
):
    """Performs sanity check on the two given inputs

    Checks if the two inputs:
        - type is Darts Timeseries
        - have the same number of components
        - if their intersection in time is not null

    Parameters
    ----------
    series_1
        1st time series
    series_2:
        2nd time series
    """

    _assert_timeseries(series_1)
    _assert_timeseries(series_2)

    # check if the two inputs time series have the same number of components
    raise_if_not(
        series_1.width == series_2.width,
        "Series must have the same number of components,"
        + f" found {series_1.width} and {series_2.width}.",
    )

    # check if the time intersection between the two inputs time series is not empty
    raise_if_not(
        len(series_1.time_index.intersection(series_2.time_index)) > 0,
        "Series must have a non-empty intersection timestamps.",
    )


def _max_pooling(series: TimeSeries, window: int) -> TimeSeries:
    """Slides a window of size `window` along the input series, and replaces the value of the
    input time series by the maximum of the values contained in the window.

    The binary time series output represents if there is an anomaly (=1) or not (=0) in the past
    window points. The new series will equal the length of the input series - window. Its first
    point will start at the first time index of the input time series + window points.

    Parameters
    ----------
    series:
        Binary time series.
    window:
        Integer value indicating the number of past samples each point represents.

    Returns
    -------
    Binary TimeSeries
    """

    raise_if_not(
        isinstance(window, int),
        f"Parameter `window` must be of type int, found {type(window)}.",
    )

    raise_if_not(
        window > 0,
        f"Parameter `window` must be stricly greater than 0, found size {window}.",
    )

    raise_if_not(
        window < len(series),
        "Parameter `window` must be smaller than the length of the input series, "
        + f" found window size {(window)}, and max size {len(series)}.",
    )

    if window == 1:
        # the process results in replacing every value by itself -> return directly the series
        return series
    else:
        return series.window_transform(
            transforms={
                "window": window,
                "function": "max",
                "mode": "rolling",
                "min_periods": window,
            },
            treat_na="dropna",
        )


def _to_list(series: Union[TimeSeries, Sequence[TimeSeries]]) -> Sequence[TimeSeries]:
    """If not already, it converts the input into a sequence

    Parameters
    ----------
    series
        single TimeSeries, or a sequence of TimeSeries

    Returns
    -------
    Sequence[TimeSeries]
    """

    return [series] if not isinstance(series, Sequence) else series


def _assert_same_length(
    list_series_1: Sequence[TimeSeries],
    list_series_2: Sequence[TimeSeries],
):
    """Checks if the two sequences contain the same number of TimeSeries."""

    raise_if_not(
        len(list_series_1) == len(list_series_2),
        "Sequences of series must be of the same length, found length:"
        + f" {len(list_series_1)} and {len(list_series_2)}.",
    )


def show_anomalies_from_scores(
    series: TimeSeries,
    model_output: TimeSeries = None,
    anomaly_scores: Union[TimeSeries, Sequence[TimeSeries]] = None,
    window: Union[int, Sequence[int]] = 1,
    names_of_scorers: Union[str, Sequence[str]] = None,
    actual_anomalies: TimeSeries = None,
    title: str = None,
    metric: str = None,
):
    """Plot the results generated by an anomaly model.

    The plot will be composed of the following:
        - the series itself with the output of the model (if given)
        - the anomaly score of each scorer. The scorer with different windows will be separated.
        - the actual anomalies, if given.

    If model_output is stochastic (i.e., if it has multiple samples), the function will plot:
        - the mean per timestamp
        - the quantile 0.95 for an upper bound
        - the quantile 0.05 for a lower bound

    Possible to:
        - add a title to the figure with the parameter `title`
        - give personalized names for the scorers with `names_of_scorers`
        - show the results of a metric for each anomaly score (AUC_ROC or AUC_PR), if the actual anomalies is given

    Parameters
    ----------
    series
        The series to visualize anomalies from.
    model_output
        Output of the model given as input the series (can be stochastic).
    anomaly_scores
        Output of the scorers given the output of the model and the series.
    window
        Window parameter for each anomaly scores.
        Default: 1. If a list of anomaly scores is given, the same default window will be used for every score.
    names_of_scorers
        Name of the scores. Must be a list of length equal to the number of scorers in the anomaly_model.
    actual_anomalies
        The ground truth of the anomalies (1 if it is an anomaly and 0 if not)
    title
        Title of the figure
    metric
        Optionally, Scoring function to use. Must be one of "AUC_ROC" and "AUC_PR".
        Default: "AUC_ROC"
    """

    raise_if_not(
        isinstance(series, TimeSeries),
        f"Input `series` must be of type TimeSeries, found {type(series)}.",
    )

    if title is None:
        if anomaly_scores is not None:
            title = "Anomaly results"
    else:
        raise_if_not(
            isinstance(title, str),
            f"Input `title` must be of type str, found {type(title)}.",
        )

    nbr_plots = 1

    if model_output is not None:
        raise_if_not(
            isinstance(model_output, TimeSeries),
            f"Input `model_output` must be of type TimeSeries, found {type(model_output)}.",
        )

    if actual_anomalies is not None:
        raise_if_not(
            isinstance(actual_anomalies, TimeSeries),
            f"Input `actual_anomalies` must be of type TimeSeries, found {type(actual_anomalies)}.",
        )

        nbr_plots = nbr_plots + 1
    else:
        raise_if_not(
            metric is None,
            "`actual_anomalies` must be given in order to calculate a metric.",
        )

    if anomaly_scores is not None:

        if isinstance(anomaly_scores, Sequence):
            for idx, s in enumerate(anomaly_scores):
                raise_if_not(
                    isinstance(s, TimeSeries),
                    f"Elements of anomaly_scores must be of type TimeSeries, found {type(s)} at index {idx}.",
                )
        else:
            raise_if_not(
                isinstance(anomaly_scores, TimeSeries),
                f"Input `anomaly_scores` must be of type TimeSeries or Sequence, found {type(actual_anomalies)}.",
            )
            anomaly_scores = [anomaly_scores]

        if names_of_scorers is not None:

            if isinstance(names_of_scorers, str):
                names_of_scorers = [names_of_scorers]
            elif isinstance(names_of_scorers, Sequence):
                for idx, name in enumerate(names_of_scorers):
                    raise_if_not(
                        isinstance(name, str),
                        f"Elements of names_of_scorers must be of type str, found {type(name)} at index {idx}.",
                    )
            else:
                raise ValueError(
                    f"Input `names_of_scorers` must be of type str or Sequence, found {type(names_of_scorers)}."
                )

            raise_if_not(
                len(names_of_scorers) == len(anomaly_scores),
                "The number of names in `names_of_scorers` must match the number of anomaly score "
                + f"given as input, found {len(names_of_scorers)} and expected {len(anomaly_scores)}.",
            )

        if isinstance(window, int):
            window = [window]
        elif isinstance(window, Sequence):
            for idx, w in enumerate(window):
                raise_if_not(
                    isinstance(w, int),
                    f"Every window must be of type int, found {type(w)} at index {idx}.",
                )
        else:
            raise ValueError(
                f"Input `window` must be of type int or Sequence, found {type(window)}."
            )

        raise_if_not(
            all([w > 0 for w in window]),
            "All windows must be positive integer.",
        )

        if len(window) == 1:
            window = window * len(anomaly_scores)
        else:
            raise_if_not(
                len(window) == len(anomaly_scores),
                "The number of window in `window` must match the number of anomaly score given as input. One "
                + f"window value for each series. Found length {len(window)}, and expected {len(anomaly_scores)}.",
            )

        raise_if_not(
            all([w < len(s) for (w, s) in zip(window, anomaly_scores)]),
            "All windows must be smaller than the length of their corresponding score.",
        )

        nbr_plots = nbr_plots + len(set(window))
    else:
        if window is not None:
            logger.warning(
                "The parameter `window` is given, but the input `anomaly_scores` is None."
            )

        if names_of_scorers is not None:
            logger.warning(
                "The parameter `names_of_scorers` is given, but the input `anomaly_scores` is None."
            )

        if metric is not None:
            logger.warning(
                "The parameter `metric` is given, but the input `anomaly_scores` is None."
            )

    fig, axs = plt.subplots(
        nbr_plots,
        figsize=(8, 4 + 2 * (nbr_plots - 1)),
        sharex=True,
        gridspec_kw={"height_ratios": [2] + [1] * (nbr_plots - 1)},
        squeeze=False,
    )

    index_ax = 0

    _plot_series(series=series, ax_id=axs[index_ax][0], linewidth=0.5, label_name="")

    if model_output is not None:

        _plot_series(
            series=model_output,
            ax_id=axs[index_ax][0],
            linewidth=0.5,
            label_name="model output",
        )

    axs[index_ax][0].set_title("")

    if actual_anomalies is not None or anomaly_scores is not None:
        axs[index_ax][0].set_xlabel("")

    axs[index_ax][0].legend(loc="upper center", bbox_to_anchor=(0.5, 1.1), ncol=2)

    if anomaly_scores is not None:

        dict_input = {}

        for idx, (score, w) in enumerate(zip(anomaly_scores, window)):

            dict_input[idx] = {"series_score": score, "window": w, "name_id": idx}

        current_window = window[0]
        index_ax = index_ax + 1

        for elem in sorted(dict_input.items(), key=lambda x: x[1]["window"]):

            idx = elem[1]["name_id"]
            w = elem[1]["window"]

            if w != current_window:
                current_window = w
                index_ax = index_ax + 1

            if metric is not None:
                value = round(
                    eval_accuracy_from_scores(
                        anomaly_score=anomaly_scores[idx],
                        actual_anomalies=actual_anomalies,
                        window=w,
                        metric=metric,
                    ),
                    3,
                )
            else:
                value = None

            if names_of_scorers is not None:
                label = names_of_scorers[idx] + [f" ({value})", ""][value is None]
            else:
                label = f"score_{str(idx)}" + [f" ({value})", ""][value is None]

            _plot_series(
                series=elem[1]["series_score"],
                ax_id=axs[index_ax][0],
                linewidth=0.5,
                label_name=label,
            )

            axs[index_ax][0].legend(
                loc="upper center", bbox_to_anchor=(0.5, 1.19), ncol=2
            )
            axs[index_ax][0].set_title(f"Window: {str(w)}", loc="left")
            axs[index_ax][0].set_title("")
            axs[index_ax][0].set_xlabel("")

    if actual_anomalies is not None:

        _plot_series(
            series=actual_anomalies,
            ax_id=axs[index_ax + 1][0],
            linewidth=1,
            label_name="anomalies",
            color="red",
        )

        axs[index_ax + 1][0].set_title("")
        axs[index_ax + 1][0].set_ylim([-0.1, 1.1])
        axs[index_ax + 1][0].set_yticks([0, 1])
        axs[index_ax + 1][0].set_yticklabels(["no", "yes"])
        axs[index_ax + 1][0].legend(
            loc="upper center", bbox_to_anchor=(0.5, 1.2), ncol=2
        )
    else:
        axs[index_ax][0].set_xlabel("timestamp")

    fig.suptitle(title)


def _plot_series(series, ax_id, linewidth, label_name, **kwargs):
    """Internal function called by ``show_anomalies_from_scores()``

    Plot the series on the given axes ax_id.

    Parameters
    ----------
    series
        The series to plot.
    ax_id
        The axis the series will be ploted on.
    linewidth
        Thickness of the line.
    label_name
        Name that will appear in the legend.
    """

    for i, c in enumerate(series._xa.component[:10]):
        comp = series._xa.sel(component=c)

        if comp.sample.size > 1:
            central_series = comp.mean(dim="sample")
            low_series = comp.quantile(q=0.05, dim="sample")
            high_series = comp.quantile(q=0.95, dim="sample")
        else:
            central_series = comp

        label_to_use = (
            (label_name + ("_" + str(i) if len(series.components) > 1 else ""))
            if label_name != ""
            else "" + str(str(c.values))
        )

        central_series.plot(ax=ax_id, linewidth=linewidth, label=label_to_use, **kwargs)

        if comp.sample.size > 1:
            ax_id.fill_between(
                series.time_index, low_series, high_series, alpha=0.25, **kwargs
            )



import logging
import shutil
import time
import unittest

# Print something for all tests taking longer than this
DURATION_THRESHOLD = 2.0


class DartsBaseTestClass(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.tic = time.time()
        logging.disable(logging.CRITICAL)

    def tearDown(self):
        duration = time.time() - self.tic
        if duration >= DURATION_THRESHOLD:
            print(f"Test {self.id()} finished after {duration:.2f} s.")
        try:
            shutil.rmtree(".darts")
        except FileNotFoundError:
            pass



import itertools
import sys
import unittest

import numpy as np
import pandas as pd

from darts import TimeSeries
from darts.dataprocessing.pipeline import Pipeline
from darts.dataprocessing.transformers import Mapper, WindowTransformer


class TimeSeriesWindowTransformTestCase(unittest.TestCase):

    times = pd.date_range("20130101", "20130110")
    series_from_values = TimeSeries.from_values(
        np.array([1, 2, 3, 4, 5, 6, 7, 8, 9, 10])
    )
    target = TimeSeries.from_times_and_values(times, range(1, 11))

    series_multi_prob = (
        (target + 10)
        .stack(target + 20)
        .concatenate((target + 100).stack(target + 200), axis=2)
    )  # 2 comps, 2 samples
    series_multi_det = (
        (target + 10).stack(target + 20).stack((target + 30).stack(target + 40))
    )  # 4 comps, 1 sample
    series_univ_det = target + 50  # 1 comp, 1 sample
    series_univ_prob = (target + 50).concatenate(
        target + 500, axis=2
    )  # 1 comp, 2 samples

    def test_ts_windowtransf_input_dictionary(self):
        """
        Test that the forecasting window transformer dictionary input parameter is correctly formatted
        """

        with self.assertRaises(TypeError):
            window_transformations = None  # None input
            self.series_univ_det.window_transform(transforms=window_transformations)

        with self.assertRaises(ValueError):
            window_transformations = []  # empty list
            self.series_univ_det.window_transform(transforms=window_transformations)

        with self.assertRaises(KeyError):
            window_transformations = {}  # empty dictionary
            self.series_univ_det.window_transform(transforms=window_transformations)

        with self.assertRaises(ValueError):
            window_transformations = [1, 2, 3]  # list of not dictionaries
            self.series_univ_det.window_transform(transforms=window_transformations)

        with self.assertRaises(KeyError):
            window_transformations = {"random_fn_name": "mean"}  # no 'function' key
            self.series_univ_det.window_transform(transforms=window_transformations)

        with self.assertRaises(AttributeError):
            window_transformations = {
                "function": "wild_fn"
            }  # not valid pandas built-in function
            self.series_univ_det.window_transform(transforms=window_transformations)

        with self.assertRaises(ValueError):
            window_transformations = {
                "function": 1
            }  # not valid pandas built-in function nore callable
            self.series_univ_det.window_transform(transforms=window_transformations)

        with self.assertRaises(ValueError):
            window_transformations = {"function": None}  # None function value
            self.series_univ_det.window_transform(transforms=window_transformations)

        with self.assertRaises(TypeError):
            window_transformations = {
                "function": "quantile",
                "window": [3],
            }  # not enough mandatory arguments for quantile
            self.series_univ_det.window_transform(transforms=window_transformations)

        with self.assertRaises(ValueError):
            window_transformations = {
                "function": "mean",
                "mode": "rolling",
                "window": -3,
            }  # negative window
            self.series_univ_det.window_transform(transforms=window_transformations)

        with self.assertRaises(ValueError):
            window_transformations = {
                "function": "mean",
                "mode": "rolling",
                "window": None,
            }  # None window
            self.series_univ_det.window_transform(transforms=window_transformations)

        with self.assertRaises(ValueError):
            window_transformations = {
                "function": "mean",
                "mode": "rolling",
                "window": [5],
            }  # window list
            self.series_univ_det.window_transform(transforms=window_transformations)

        # skip this test for Python <= 3.7
        if sys.version_info >= (3, 8):
            with self.assertRaises(ValueError):
                window_transformations = {
                    "function": "mean",
                    "window": 3,
                    "mode": "rolling",
                    "step": -2,
                }  # Negative step
                self.series_univ_det.window_transform(transforms=window_transformations)

        with self.assertRaises(ValueError):
            window_transformations = {
                "function": "mean",
                "window": 3,
                "mode": "rnd",
            }  # invalid mode
            self.series_univ_det.window_transform(transforms=window_transformations)

        with self.assertRaises(ValueError):
            window_transformations = {
                "function": "mean",
                "mode": "rolling",
                "window": 3,
                "center": "True",
            }  # forecating_safe=True vs center=True
            self.series_univ_det.window_transform(transforms=window_transformations)

    def test_ts_windowtransf_output_series(self):
        # univariate deterministic input
        transforms = {"function": "sum", "mode": "rolling", "window": 1}
        transformed_ts = self.series_univ_det.window_transform(transforms=transforms)

        self.assertEqual(
            list(itertools.chain(*transformed_ts.values().tolist())),
            list(itertools.chain(*self.series_univ_det.values().tolist())),
        )
        self.assertEqual(
            transformed_ts.components.to_list(),
            [
                f"{transforms['mode']}_{transforms['function']}_{str(transforms['window'])}_{comp}"
                for comp in self.series_univ_det.components
            ],
        )

        # test customized function name that overwrites the pandas builtin transformation
        transforms = {
            "function": "sum",
            "mode": "rolling",
            "window": 1,
            "function_name": "customized_name",
        }
        transformed_ts = self.series_univ_det.window_transform(transforms=transforms)
        self.assertEqual(
            transformed_ts.components.to_list(),
            [
                f"{transforms['mode']}_{transforms['function_name']}_{str(transforms['window'])}_{comp}"
                for comp in self.series_univ_det.components
            ],
        )
        del transforms["function_name"]

        # multivariate deterministic input
        # transform one component
        transforms.update({"components": "0"})

        transformed_ts = self.series_multi_det.window_transform(transforms=transforms)
        self.assertEqual(
            transformed_ts.components.to_list(),
            [
                f"{transforms['mode']}_{transforms['function']}_{str(transforms['window'])}_{comp}"
                for comp in transforms["components"]
            ],
        )

        transformed_ts = self.series_multi_det.window_transform(
            transforms=transforms, keep_non_transformed=True
        )

        self.assertEqual(
            transformed_ts.components.to_list(),
            [
                f"{transforms['mode']}_{transforms['function']}_{str(transforms['window'])}_{comp}"
                for comp in transforms["components"]
            ]
            + self.series_multi_det.components.to_list(),
        )

        # transform multiple components
        transforms = {
            "function": "sum",
            "mode": "rolling",
            "window": 1,
            "components": ["0", "0_1"],
        }

        transformed_ts = self.series_multi_det.window_transform(transforms=transforms)
        self.assertEqual(
            transformed_ts.components.to_list(),
            [
                f"{transforms['mode']}_{transforms['function']}_{str(transforms['window'])}_{comp}"
                for comp in transforms["components"]
            ],
        )

        transformed_ts = self.series_multi_det.window_transform(
            transforms=transforms, keep_non_transformed=True
        )

        self.assertEqual(
            transformed_ts.components.to_list(),
            [
                f"{transforms['mode']}_{transforms['function']}_{str(transforms['window'])}_{comp}"
                for comp in transforms["components"]
            ]
            + self.series_multi_det.components.to_list(),
        )

        # multiple transformations

        transforms = [transforms] + [
            {
                "function": "mean",
                "mode": "rolling",
                "window": 1,
                "components": ["0", "0_1"],
            }
        ]

        transformed_ts = self.series_multi_det.window_transform(transforms=transforms)
        self.assertEqual(
            transformed_ts.components.to_list(),
            [
                f"{transformation['mode']}_{transformation['function']}_{str(transformation['window'])}_{comp}"
                for transformation in transforms
                for comp in transformation["components"]
            ],
        )

        transformed_ts = self.series_multi_det.window_transform(
            transforms=transforms, keep_non_transformed=True
        )
        self.assertEqual(
            transformed_ts.components.to_list(),
            [
                f"{transformation['mode']}_{transformation['function']}_{str(transformation['window'])}_{comp}"
                for transformation in transforms
                for comp in transformation["components"]
            ]
            + self.series_multi_det.components.to_list(),
        )

        # multivariate probabilistic input
        transformed_ts = self.series_multi_prob.window_transform(transforms=transforms)
        self.assertEqual(transformed_ts.n_samples, 2)

    def test_user_defined_function_behavior(self):
        def count_above_mean(array):
            mean = np.mean(array)
            return np.where(array > mean)[0].size

        transformation = {
            "function": count_above_mean,
            "mode": "rolling",
            "window": 5,
        }
        transformed_ts = self.target.window_transform(
            transformation,
        )
        expected_transformed_series = TimeSeries.from_times_and_values(
            self.times,
            [0, 1, 1, 2, 2, 2, 2, 2, 2, 2],
            columns=["rolling_udf_5_0"],
        )
        self.assertEqual(transformed_ts, expected_transformed_series)

        # test if a customized function name is provided
        transformation.update({"function_name": "count_above_mean"})
        transformed_ts = self.target.window_transform(
            transformation,
        )
        self.assertEqual(
            transformed_ts.components.to_list(),
            [
                f"{transformation['mode']}_{transformation['function_name']}_{str(transformation['window'])}_{comp}"
                for comp in self.target.components
            ],
        )

    def test_ts_windowtransf_output_nabehavior(self):
        window_transformations = {
            "function": "sum",
            "mode": "rolling",
            "window": 3,
            "min_periods": 2,
        }

        # fill na with a specific value
        transformed_ts = self.target.window_transform(
            window_transformations, treat_na=100
        )
        expected_transformed_series = TimeSeries.from_times_and_values(
            self.times,
            [100, 3, 6, 9, 12, 15, 18, 21, 24, 27],
            columns=["rolling_sum_3_2_0"],
        )
        self.assertEqual(transformed_ts, expected_transformed_series)

        # dropna
        transformed_ts = self.target.window_transform(
            window_transformations, treat_na="dropna"
        )
        expected_transformed_series = TimeSeries.from_times_and_values(
            self.times[1:],
            [3, 6, 9, 12, 15, 18, 21, 24, 27],
            columns=["rolling_sum_3_2_0"],
        )
        self.assertEqual(transformed_ts, expected_transformed_series)

        # backfill na
        transformed_ts = self.target.window_transform(
            window_transformations, treat_na="bfill", forecasting_safe=False
        )
        # backfill works only with forecasting_safe=False
        expected_transformed_series = TimeSeries.from_times_and_values(
            self.times,
            [3, 3, 6, 9, 12, 15, 18, 21, 24, 27],
            columns=["rolling_sum_3_2_0"],
        )
        self.assertEqual(transformed_ts, expected_transformed_series)

        with self.assertRaises(ValueError):
            # uknonwn treat_na
            self.target.window_transform(
                window_transformations, treat_na="fillrnd", forecasting_safe=False
            )

        with self.assertRaises(ValueError):
            # unauhtorized treat_na=bfill with forecasting_safe=True
            self.target.window_transform(window_transformations, treat_na="bfill")

    def test_tranformed_ts_index(self):

        # DateTimeIndex
        transformed_series = self.target.window_transform({"function": "sum"})
        self.assertEqual(
            self.target._time_index.__class__, transformed_series._time_index.__class__
        )
        # length index should not change for default transformation configurations
        self.assertEqual(
            len(self.target._time_index), len(transformed_series._time_index)
        )
        # RangeIndex
        transformed_series = self.series_from_values.window_transform(
            {"function": "sum"}
        )
        self.assertEqual(
            self.series_from_values._time_index.__class__,
            transformed_series._time_index.__class__,
        )
        self.assertEqual(
            len(self.series_from_values._time_index),
            len(transformed_series._time_index),
        )

    def test_include_current(self):
        # if "closed"="left" should not shift the index
        transformation = {
            "function": "sum",
            "mode": "rolling",
            "window": 1,
            "closed": "left",
        }
        expected_transformed_series = TimeSeries.from_times_and_values(
            self.times,
            ["NaN", 1, 2, 3, 4, 5, 6, 7, 8, 9],
            columns=["rolling_sum_1_0"],
        )
        transformed_ts = self.target.window_transform(
            transformation, include_current=False
        )
        self.assertEqual(transformed_ts, expected_transformed_series)

        # shift the index
        transformation = {"function": "sum", "mode": "rolling", "window": 1}
        expected_transformed_series = TimeSeries.from_times_and_values(
            self.times,
            ["NaN", 1, 2, 3, 4, 5, 6, 7, 8, 9],
            columns=["rolling_sum_1_0"],
        )
        transformed_ts = self.target.window_transform(
            transformation, include_current=False
        )
        self.assertEqual(transformed_ts, expected_transformed_series)

        # pandas on Python <= 3.7 raises this error:
        # AttributeError: 'ExponentialMovingWindow' object has no attribute 'sum'

        # skip these tests in Python <= 3.7:
        if sys.version_info < (3, 8):
            return

        transformation = [
            {"function": "sum", "mode": "rolling", "window": 1, "closed": "left"},
            {"function": "sum", "mode": "ewm", "span": 1},
        ]
        expected_transformed_series = TimeSeries.from_times_and_values(
            self.times,
            [
                ["NaN", "NaN"],
                [1, 1],
                [2, 2],
                [3, 3],
                [4, 4],
                [5, 5],
                [6, 6],
                [7, 7],
                [8, 8],
                [9, 9],
            ],
            columns=["rolling_sum_1_0", "ewm_sum_0"],
        )
        transformed_ts = self.target.window_transform(
            transformation, include_current=False
        )
        self.assertEqual(transformed_ts, expected_transformed_series)

        expected_transformed_series = TimeSeries.from_times_and_values(
            self.times,
            [
                [1, 1],
                [1, 1],
                [2, 2],
                [3, 3],
                [4, 4],
                [5, 5],
                [6, 6],
                [7, 7],
                [8, 8],
                [9, 9],
            ],
            columns=["rolling_sum_1_0", "ewm_sum_0"],
        )
        transformed_ts = self.target.window_transform(
            transformation,
            include_current=False,
            forecasting_safe=False,
            treat_na="bfill",
        )
        self.assertEqual(transformed_ts, expected_transformed_series)

        transformation = [
            {
                "function": "sum",
                "mode": "rolling",
                "window": 2,
                "closed": "left",
                "min_periods": 2,
            },
            {"function": "sum", "mode": "ewm", "span": 1, "min_periods": 2},
        ]

        expected_transformed_series = TimeSeries.from_times_and_values(
            self.times,
            [
                ["NaN", "NaN"],
                ["NaN", "NaN"],
                [3, 2],
                [5, 3],
                [7, 4],
                [9, 5],
                [11, 6],
                [13, 7],
                [15, 8],
                [17, 9],
            ],
            columns=["rolling_sum_2_2_0", "ewm_sum_2_0"],
        )

        transformed_ts = self.target.window_transform(
            transformation, include_current=False
        )
        self.assertEqual(transformed_ts, expected_transformed_series)


class WindowTransformerTestCase(unittest.TestCase):

    times = pd.date_range("20130101", "20130110")
    target = TimeSeries.from_times_and_values(times, range(1, 11))
    times_hourly = pd.date_range(start="20130101", freq="1H", periods=10)
    target_hourly = TimeSeries.from_times_and_values(times_hourly, range(1, 11))

    series_multi_prob = (
        (target + 10)
        .stack(target + 20)
        .concatenate((target + 100).stack(target + 200), axis=2)
    )  # 2 comps, 2 samples
    series_multi_det = (
        (target + 10).stack(target + 20).stack((target + 30).stack(target + 40))
    )  # 4 comps, 1 sample
    series_univ_det = target + 50  # 1 comp, 1 sample
    series_univ_prob = (target + 50).concatenate(
        target + 500, axis=2
    )  # 1 comp, 2 samples

    sequence_det = [series_univ_det, series_multi_det]
    sequence_prob = [series_univ_prob, series_multi_prob]

    def test_window_transformer_output(self):
        window_transformations = {
            "function": "sum",
            "components": ["0"],
        }
        transformer = WindowTransformer(
            transforms=window_transformations,
            treat_na=100,
            keep_non_transformed=True,
            forecasting_safe=True,
        )
        transformed_ts_list = transformer.transform(self.sequence_det)

        self.assertEqual(len(transformed_ts_list), 2)
        self.assertEqual(transformed_ts_list[0].n_components, 2)
        self.assertEqual(
            transformed_ts_list[0].n_timesteps, self.series_multi_det.n_timesteps
        )
        self.assertEqual(transformed_ts_list[1].n_components, 5)
        self.assertEqual(
            transformed_ts_list[1].n_timesteps, self.series_multi_det.n_timesteps
        )

    def test_window_transformer_offset_parameter(self):
        """
        Test that the window parameter can support offset of pandas.Timedelta
        """
        base_parameters = {
            "function": "mean",
            "components": ["0"],
            "mode": "rolling",
        }

        offset_parameters = base_parameters.copy()
        offset_parameters.update({"window": pd.Timedelta(hours=4)})
        offset_transformer = WindowTransformer(
            transforms=offset_parameters,
        )
        offset_transformed = offset_transformer.transform(self.target_hourly)

        integer_parameters = base_parameters.copy()
        integer_parameters.update({"window": 4})
        integer_transformer = WindowTransformer(
            transforms=integer_parameters,
        )
        integer_transformed = integer_transformer.transform(self.target_hourly)
        np.testing.assert_equal(
            integer_transformed.values(), offset_transformed.values()
        )
        self.assertEqual(
            offset_transformed.components[0], "rolling_mean_0 days 04:00:00_0"
        )
        self.assertEqual(integer_transformed.components[0], "rolling_mean_4_0")

        invalid_parameters = base_parameters.copy()
        invalid_parameters.update({"window": pd.DateOffset(hours=4)})
        invalid_transformer = WindowTransformer(
            transforms=invalid_parameters,
        )
        # if pd.DateOffset, raise ValueError of non-fixed frequency
        with self.assertRaises(ValueError):
            invalid_transformer.transform(self.target_hourly)

    def test_transformers_pipeline(self):
        """
        Test that the forecasting window transformer can be used in a pipeline

        """

        times1 = pd.date_range("20130101", "20130110")
        series_1 = TimeSeries.from_times_and_values(times1, range(1, 11))

        expected_transformed_series = TimeSeries.from_times_and_values(
            times1,
            [100, 15, 30, 45, 60, 75, 90, 105, 120, 135],
            columns=["rolling_sum_3_2_0"],
        )

        # adds NaNs
        window_transformations = [
            {"function": "sum", "mode": "rolling", "window": 3, "min_periods": 2}
        ]

        def times_five(x):
            return x * 5

        mapper = Mapper(fn=times_five)

        window_transformer = WindowTransformer(
            transforms=window_transformations, treat_na=100
        )

        pipeline = Pipeline([mapper, window_transformer])

        transformed_series = pipeline.fit_transform(series_1)

        self.assertEqual(transformed_series, expected_transformed_series)



"""
Linear Regression model
-----------------------

A forecasting model using a linear regression of some of the target series' lags, as well as optionally some
covariate series lags in order to obtain a forecast.
"""
from typing import List, Optional, Sequence, Tuple, Union

import numpy as np
from scipy.optimize import linprog
from sklearn.linear_model import LinearRegression, PoissonRegressor, QuantileRegressor

from darts.logging import get_logger
from darts.models.forecasting.regression_model import RegressionModel, _LikelihoodMixin
from darts.timeseries import TimeSeries

logger = get_logger(__name__)


class LinearRegressionModel(RegressionModel, _LikelihoodMixin):
    def __init__(
        self,
        lags: Union[int, list] = None,
        lags_past_covariates: Union[int, List[int]] = None,
        lags_future_covariates: Union[Tuple[int, int], List[int]] = None,
        output_chunk_length: int = 1,
        add_encoders: Optional[dict] = None,
        likelihood: str = None,
        quantiles: List[float] = None,
        random_state: Optional[int] = None,
        multi_models: Optional[bool] = True,
        use_static_covariates: bool = True,
        **kwargs,
    ):
        """Linear regression model.

        Parameters
        ----------
        lags
            Lagged target values used to predict the next time step. If an integer is given the last `lags` past lags
            are used (from -1 backward). Otherwise a list of integers with lags is required (each lag must be < 0).
        lags_past_covariates
            Number of lagged past_covariates values used to predict the next time step. If an integer is given the last
            `lags_past_covariates` past lags are used (inclusive, starting from lag -1). Otherwise a list of integers
            with lags < 0 is required.
        lags_future_covariates
            Number of lagged future_covariates values used to predict the next time step. If an tuple (past, future) is
            given the last `past` lags in the past are used (inclusive, starting from lag -1) along with the first
            `future` future lags (starting from 0 - the prediction time - up to `future - 1` included). Otherwise a list
            of integers with lags is required.
        output_chunk_length
            Number of time steps predicted at once by the internal regression model. Does not have to equal the forecast
            horizon `n` used in `predict()`. However, setting `output_chunk_length` equal to the forecast horizon may
            be useful if the covariates don't extend far enough into the future.
        add_encoders
            A large number of past and future covariates can be automatically generated with `add_encoders`.
            This can be done by adding multiple pre-defined index encoders and/or custom user-made functions that
            will be used as index encoders. Additionally, a transformer such as Darts' :class:`Scaler` can be added to
            transform the generated covariates. This happens all under one hood and only needs to be specified at
            model creation.
            Read :meth:`SequentialEncoder <darts.dataprocessing.encoders.SequentialEncoder>` to find out more about
            ``add_encoders``. Default: ``None``. An example showing some of ``add_encoders`` features:

            .. highlight:: python
            .. code-block:: python

                add_encoders={
                    'cyclic': {'future': ['month']},
                    'datetime_attribute': {'future': ['hour', 'dayofweek']},
                    'position': {'past': ['relative'], 'future': ['relative']},
                    'custom': {'past': [lambda idx: (idx.year - 1950) / 50]},
                    'transformer': Scaler()
                }
            ..
        likelihood
            Can be set to `quantile` or `poisson`. If set, the model will be probabilistic, allowing sampling at
            prediction time. If set to `quantile`, the `sklearn.linear_model.QuantileRegressor` is used. Similarly, if
            set to `poisson`, the `sklearn.linear_model.PoissonRegressor` is used.
        quantiles
            Fit the model to these quantiles if the `likelihood` is set to `quantile`.
        random_state
            Control the randomness of the sampling. Used as seed for
            `numpy.random.Generator
            <https://numpy.org/doc/stable/reference/random/generator.html#numpy.random.Generator>`_. Ignored when
            no `likelihood` is set.
            Default: ``None``.
        multi_models
            If True, a separate model will be trained for each future lag to predict. If False, a single model is
            trained to predict at step 'output_chunk_length' in the future. Default: True.
        use_static_covariates
            Whether the model should use static covariate information in case the input `series` passed to ``fit()``
            contain static covariates. If ``True``, and static covariates are available at fitting time, will enforce
            that all target `series` have the same static covariate dimensionality in ``fit()`` and ``predict()`.
        **kwargs
            Additional keyword arguments passed to `sklearn.linear_model.LinearRegression` (by default), to
            `sklearn.linear_model.PoissonRegressor` (if `likelihood="poisson"`), or to
            `sklearn.linear_model.QuantileRegressor` (if `likelihood="quantile"`).
        """
        self.kwargs = kwargs
        self._median_idx = None
        self._model_container = None
        self.quantiles = None
        self.likelihood = likelihood
        self._rng = None

        # parse likelihood
        available_likelihoods = ["quantile", "poisson"]  # to be extended
        if likelihood is not None:
            self._check_likelihood(likelihood, available_likelihoods)
            self._rng = np.random.default_rng(seed=random_state)

            if likelihood == "poisson":
                model = PoissonRegressor(**kwargs)
            if likelihood == "quantile":
                model = QuantileRegressor(**kwargs)
                self.quantiles, self._median_idx = self._prepare_quantiles(quantiles)
                self._model_container = self._get_model_container()
        else:
            model = LinearRegression(**kwargs)

        super().__init__(
            lags=lags,
            lags_past_covariates=lags_past_covariates,
            lags_future_covariates=lags_future_covariates,
            output_chunk_length=output_chunk_length,
            add_encoders=add_encoders,
            model=model,
            multi_models=multi_models,
            use_static_covariates=use_static_covariates,
        )

    def fit(
        self,
        series: Union[TimeSeries, Sequence[TimeSeries]],
        past_covariates: Optional[Union[TimeSeries, Sequence[TimeSeries]]] = None,
        future_covariates: Optional[Union[TimeSeries, Sequence[TimeSeries]]] = None,
        max_samples_per_ts: Optional[int] = None,
        n_jobs_multioutput_wrapper: Optional[int] = None,
        **kwargs,
    ):
        """
        Fit/train the model on one or multiple series.

        Parameters
        ----------
        series
            TimeSeries or Sequence[TimeSeries] object containing the target values.
        past_covariates
            Optionally, a series or sequence of series specifying past-observed covariates
        future_covariates
            Optionally, a series or sequence of series specifying future-known covariates
        max_samples_per_ts
            This is an integer upper bound on the number of tuples that can be produced
            per time series. It can be used in order to have an upper bound on the total size of the dataset and
            ensure proper sampling. If `None`, it will read all of the individual time series in advance (at dataset
            creation) to know their sizes, which might be expensive on big datasets.
            If some series turn out to have a length that would allow more than `max_samples_per_ts`, only the
            most recent `max_samples_per_ts` samples will be considered.
        n_jobs_multioutput_wrapper
            Number of jobs of the MultiOutputRegressor wrapper to run in parallel. Only used if the model doesn't
            support multi-output regression natively.
        **kwargs
            Additional keyword arguments passed to the `fit` method of the model.
        """

        if self.likelihood == "quantile":
            # set solver for linear program
            if "solver" not in self.kwargs:
                # set default fast solver
                self.kwargs["solver"] = "highs"

            # test solver availability with dummy problem
            c = [1]
            try:
                linprog(c=c, method=self.kwargs["solver"])
            except ValueError as ve:
                logger.warning(
                    f"{ve}. Upgrading scipy enables significantly faster solvers"
                )
                # set solver to slow legacy
                self.kwargs["solver"] = "interior-point"

            # empty model container in case of multiple calls to fit, e.g. when backtesting
            self._model_container.clear()

            for quantile in self.quantiles:
                self.kwargs["quantile"] = quantile
                self.model = QuantileRegressor(**self.kwargs)
                super().fit(
                    series=series,
                    past_covariates=past_covariates,
                    future_covariates=future_covariates,
                    max_samples_per_ts=max_samples_per_ts,
                    **kwargs,
                )

                self._model_container[quantile] = self.model

            return self

        else:
            super().fit(
                series=series,
                past_covariates=past_covariates,
                future_covariates=future_covariates,
                max_samples_per_ts=max_samples_per_ts,
                **kwargs,
            )

            return self

    def _predict_and_sample(
        self, x: np.ndarray, num_samples: int, **kwargs
    ) -> np.ndarray:
        if self.likelihood == "quantile":
            return self._predict_quantiles(x, num_samples, **kwargs)
        elif self.likelihood == "poisson":
            return self._predict_poisson(x, num_samples, **kwargs)
        else:
            return super()._predict_and_sample(x, num_samples, **kwargs)

    def _is_probabilistic(self) -> bool:
        return self.likelihood is not None



"""
NLL Poisson Scorer
------------------

Poisson distribution negative log-likelihood Scorer.

The anomaly score is the negative log likelihood of the actual series values
under a Poisson distribution estimated from the stochastic prediction.
"""

import numpy as np
from scipy.stats import poisson

from darts.ad.scorers.scorers import NLLScorer


class PoissonNLLScorer(NLLScorer):
    def __init__(self, window: int = 1) -> None:
        super().__init__(window=window)

    def __str__(self):
        return "PoissonNLLScorer"

    def _score_core_nllikelihood(
        self,
        deterministic_values: np.ndarray,
        probabilistic_estimations: np.ndarray,
    ) -> np.ndarray:

        mu = np.mean(probabilistic_estimations, axis=1)
        return -poisson.logpmf(deterministic_values, mu=mu)



import random

import numpy as np

from darts.metrics import mape
from darts.models import FourTheta, Theta
from darts.tests.base_test_class import DartsBaseTestClass
from darts.utils.timeseries_generation import linear_timeseries as lt
from darts.utils.timeseries_generation import random_walk_timeseries as rt
from darts.utils.timeseries_generation import sine_timeseries as st
from darts.utils.utils import ModelMode, SeasonalityMode, TrendMode


class FourThetaTestCase(DartsBaseTestClass):
    def test_input(self):
        with self.assertRaises(ValueError):
            FourTheta(model_mode=SeasonalityMode.ADDITIVE)
        with self.assertRaises(ValueError):
            FourTheta(season_mode=ModelMode.ADDITIVE)
        with self.assertRaises((ValueError, TypeError)):
            FourTheta(trend_mode="linear")

    def test_negative_series(self):
        sine_series = st(length=50)
        model = FourTheta(
            model_mode=ModelMode.MULTIPLICATIVE,
            trend_mode=TrendMode.EXPONENTIAL,
            season_mode=SeasonalityMode.ADDITIVE,
            normalization=False,
        )
        model.fit(sine_series)
        self.assertTrue(
            model.model_mode is ModelMode.ADDITIVE
            and model.trend_mode is TrendMode.LINEAR
        )

    def test_zero_mean(self):
        sine_series = st(length=50)
        with self.assertRaises(ValueError):
            model = FourTheta(
                model_mode=ModelMode.MULTIPLICATIVE, trend_mode=TrendMode.EXPONENTIAL
            )
            model.fit(sine_series)

    def test_theta(self):
        np.random.seed(1)
        series = rt(length=50, mean=100)
        theta_param = np.random.randint(1, 5)
        theta = Theta(theta_param)
        fourtheta = FourTheta(theta_param, normalization=False)
        theta.fit(series)
        fourtheta.fit(series)
        forecast_theta = theta.predict(20)
        forecast_fourtheta = fourtheta.predict(20)
        weighted_delta = (forecast_theta - forecast_fourtheta) / forecast_theta
        self.assertTrue((weighted_delta <= 3e-5).all().item())

    def test_best_model(self):
        random.seed(1)
        sine_series = st(length=50, value_y_offset=50)
        linear_series = lt(length=50)
        series = sine_series + linear_series
        train_series, val_series = series.split_before(series.time_index[-10])
        thetas = np.linspace(-3, 3, 30)
        best_model, _, _ = FourTheta.select_best_model(train_series, thetas)
        model = FourTheta(
            random.choice(thetas),
            model_mode=random.choice(list(ModelMode)),
            trend_mode=random.choice(list(TrendMode)),
            season_mode=random.choice(list(SeasonalityMode)),
        )
        model.fit(train_series)
        best_model.fit(train_series)
        forecast_random = model.predict(10)
        forecast_best = best_model.predict(10)
        self.assertTrue(
            mape(val_series, forecast_best) <= mape(val_series, forecast_random)
        )

    def test_min_train_series_length_with_seasonality(self):
        seasonality_period = 12
        fourtheta = FourTheta(
            model_mode=ModelMode.MULTIPLICATIVE,
            trend_mode=TrendMode.EXPONENTIAL,
            season_mode=SeasonalityMode.ADDITIVE,
            seasonality_period=seasonality_period,
            normalization=False,
        )
        theta = Theta(
            season_mode=SeasonalityMode.ADDITIVE,
            seasonality_period=seasonality_period,
        )
        self.assertEqual(fourtheta.min_train_series_length, 2 * seasonality_period)
        self.assertEqual(theta.min_train_series_length, 2 * seasonality_period)

    def test_min_train_series_length_without_seasonality(self):
        fourtheta = FourTheta(
            model_mode=ModelMode.MULTIPLICATIVE,
            trend_mode=TrendMode.EXPONENTIAL,
            season_mode=SeasonalityMode.ADDITIVE,
            seasonality_period=None,
            normalization=False,
        )
        theta = Theta(
            season_mode=SeasonalityMode.ADDITIVE,
            seasonality_period=None,
        )
        self.assertEqual(fourtheta.min_train_series_length, 3)
        self.assertEqual(theta.min_train_series_length, 3)

    def test_fit_insufficient_train_series_length(self):
        sine_series = st(length=21, freq="MS")
        with self.assertRaises(ValueError):
            fourtheta = FourTheta(
                model_mode=ModelMode.MULTIPLICATIVE,
                trend_mode=TrendMode.EXPONENTIAL,
                season_mode=SeasonalityMode.ADDITIVE,
                seasonality_period=12,
            )
            fourtheta.fit(sine_series)
        with self.assertRaises(ValueError):
            theta = Theta(
                season_mode=SeasonalityMode.ADDITIVE,
                seasonality_period=12,
            )
            theta.fit(sine_series)



import copy
import os
import shutil
import tempfile
from typing import Callable
from unittest.mock import Mock, patch

import numpy as np
import pandas as pd
import pytest

from darts.datasets import AirPassengersDataset, IceCreamHeaterDataset
from darts.logging import get_logger
from darts.metrics import mape
from darts.models import (
    ARIMA,
    BATS,
    FFT,
    TBATS,
    VARIMA,
    AutoARIMA,
    Croston,
    ExponentialSmoothing,
    FourTheta,
    KalmanForecaster,
    LinearRegressionModel,
    NaiveDrift,
    NaiveMean,
    NaiveMovingAverage,
    NaiveSeasonal,
    Prophet,
    RandomForest,
    RegressionModel,
    StatsForecastAutoARIMA,
    StatsForecastAutoCES,
    StatsForecastAutoETS,
    StatsForecastAutoTheta,
    Theta,
)
from darts.models.forecasting.forecasting_model import (
    LocalForecastingModel,
    TransferableFutureCovariatesLocalForecastingModel,
)
from darts.tests.base_test_class import DartsBaseTestClass
from darts.timeseries import TimeSeries
from darts.utils import timeseries_generation as tg
from darts.utils.utils import ModelMode, SeasonalityMode, TrendMode

logger = get_logger(__name__)

# (forecasting models, maximum error) tuples
models = [
    (ExponentialSmoothing(), 5.4),
    (ARIMA(12, 2, 1), 5.2),
    (ARIMA(1, 1, 1), 24),
    (StatsForecastAutoARIMA(season_length=12), 4.6),
    (StatsForecastAutoTheta(season_length=12), 5.5),
    (StatsForecastAutoCES(season_length=12, model="Z"), 7.3),
    (StatsForecastAutoETS(season_length=12, model="AAZ"), 4.1),
    (Croston(version="classic"), 23),
    (Croston(version="tsb", alpha_d=0.1, alpha_p=0.1), 23),
    (Theta(), 11),
    (Theta(1), 17),
    (Theta(-1), 12),
    (FourTheta(1), 17),
    (FourTheta(-1), 12),
    (FourTheta(trend_mode=TrendMode.EXPONENTIAL), 6.0),
    (FourTheta(model_mode=ModelMode.MULTIPLICATIVE), 10),
    (FourTheta(season_mode=SeasonalityMode.ADDITIVE), 12),
    (FFT(trend="poly"), 13),
    (KalmanForecaster(dim_x=3), 20),
    (LinearRegressionModel(lags=12), 13),
    (RandomForest(lags=12, n_estimators=5, max_depth=3), 14),
    (Prophet(), 9.0),
    (AutoARIMA(), 12),
    (TBATS(use_trend=True, use_arma_errors=True, use_box_cox=True), 8.5),
    (BATS(use_trend=True, use_arma_errors=True, use_box_cox=True), 11),
]

# forecasting models with exogenous variables support
multivariate_models = [
    (VARIMA(1, 0, 0), 32),
    (VARIMA(1, 1, 1), 25),
    (KalmanForecaster(dim_x=30), 16),
    (NaiveSeasonal(), 32),
    (NaiveMean(), 37),
    (NaiveDrift(), 39),
    (NaiveMovingAverage(input_chunk_length=5), 34),
]

dual_models = [
    ARIMA(),
    StatsForecastAutoARIMA(season_length=12),
    StatsForecastAutoETS(season_length=12),
    Prophet(),
    AutoARIMA(),
]

# test only a few models for encoder support reduce time
encoder_support_models = [
    VARIMA(1, 0, 0),
    ARIMA(),
    AutoARIMA(),
    Prophet(),
    KalmanForecaster(dim_x=30),
]


class LocalForecastingModelsTestCase(DartsBaseTestClass):

    # forecasting horizon used in runnability tests
    forecasting_horizon = 5

    # dummy timeseries for runnability tests
    np.random.seed(1)
    ts_gaussian = tg.gaussian_timeseries(length=100, mean=50)
    # for testing covariate slicing
    ts_gaussian_long = tg.gaussian_timeseries(
        length=len(ts_gaussian) + 2 * forecasting_horizon,
        start=ts_gaussian.start_time() - forecasting_horizon * ts_gaussian.freq,
        mean=50,
    )

    # real timeseries for functionality tests
    ts_passengers = AirPassengersDataset().load()
    ts_pass_train, ts_pass_val = ts_passengers.split_after(pd.Timestamp("19570101"))

    # real multivariate timeseries for functionality tests
    ts_ice_heater = IceCreamHeaterDataset().load()
    ts_ice_heater_train, ts_ice_heater_val = ts_ice_heater.split_after(split_point=0.7)

    def retrain_func(
        counter, pred_time, train_series, past_covariates, future_covariates
    ):
        return len(train_series) % 2 == 0

    def setUp(self):
        self.temp_work_dir = tempfile.mkdtemp(prefix="darts")

    def tearDown(self):
        shutil.rmtree(self.temp_work_dir)

    def test_save_model_parameters(self):
        # model creation parameters were saved before. check if re-created model has same params as original
        for model, _ in models:
            self.assertTrue(
                model._model_params == model.untrained_model()._model_params
            )

    def test_save_load_model(self):
        # check if save and load methods work and if loaded model creates same forecasts as original model
        cwd = os.getcwd()
        os.chdir(self.temp_work_dir)

        for model in [ARIMA(1, 1, 1), LinearRegressionModel(lags=12)]:
            model_path_str = type(model).__name__
            model_path_file = model_path_str + "_file"
            model_paths = [model_path_str, model_path_file]
            full_model_paths = [
                os.path.join(self.temp_work_dir, p) for p in model_paths
            ]

            model.fit(self.ts_gaussian)
            model_prediction = model.predict(self.forecasting_horizon)

            # test save
            model.save()
            model.save(model_path_str)
            with open(model_path_file, "wb") as f:
                model.save(f)

            for p in full_model_paths:
                self.assertTrue(os.path.exists(p))

            self.assertTrue(
                len(
                    [
                        p
                        for p in os.listdir(self.temp_work_dir)
                        if p.startswith(type(model).__name__)
                    ]
                )
                == 3
            )

            # test load
            loaded_model_str = type(model).load(model_path_str)
            loaded_model_file = type(model).load(model_path_file)
            loaded_models = [loaded_model_str, loaded_model_file]

            for loaded_model in loaded_models:
                self.assertEqual(
                    model_prediction, loaded_model.predict(self.forecasting_horizon)
                )

        os.chdir(cwd)

    def test_models_runnability(self):
        for model, _ in models:
            if not isinstance(model, RegressionModel):
                self.assertTrue(isinstance(model, LocalForecastingModel))
            prediction = model.fit(self.ts_gaussian).predict(self.forecasting_horizon)
            self.assertTrue(len(prediction) == self.forecasting_horizon)

    def test_models_performance(self):
        # for every model, check whether its errors do not exceed the given bounds
        for model, max_mape in models:
            np.random.seed(1)  # some models are probabilist...
            model.fit(self.ts_pass_train)
            prediction = model.predict(len(self.ts_pass_val))
            current_mape = mape(self.ts_pass_val, prediction)
            self.assertTrue(
                current_mape < max_mape,
                "{} model exceeded the maximum MAPE of {}. "
                "with a MAPE of {}".format(str(model), max_mape, current_mape),
            )

    def test_multivariate_models_performance(self):
        # for every model, check whether its errors do not exceed the given bounds
        for model, max_mape in multivariate_models:
            np.random.seed(1)
            model.fit(self.ts_ice_heater_train)
            prediction = model.predict(len(self.ts_ice_heater_val))
            current_mape = mape(self.ts_ice_heater_val, prediction)
            self.assertTrue(
                current_mape < max_mape,
                "{} model exceeded the maximum MAPE of {}. "
                "with a MAPE of {}".format(str(model), max_mape, current_mape),
            )

    def test_multivariate_input(self):
        es_model = ExponentialSmoothing()
        ts_passengers_enhanced = self.ts_passengers.add_datetime_attribute("month")
        with self.assertRaises(ValueError):
            es_model.fit(ts_passengers_enhanced)
        es_model.fit(ts_passengers_enhanced["#Passengers"])
        with self.assertRaises(KeyError):
            es_model.fit(ts_passengers_enhanced["2"])

    def test_exogenous_variables_support(self):
        # test case with pd.DatetimeIndex
        target_dt_idx = self.ts_gaussian
        fc_dt_idx = self.ts_gaussian_long

        # test case with numerical pd.RangeIndex
        target_num_idx = TimeSeries.from_times_and_values(
            times=tg.generate_index(start=0, length=len(self.ts_gaussian)),
            values=self.ts_gaussian.all_values(copy=False),
        )
        fc_num_idx = TimeSeries.from_times_and_values(
            times=tg.generate_index(start=0, length=len(self.ts_gaussian_long)),
            values=self.ts_gaussian_long.all_values(copy=False),
        )

        for target, future_covariates in zip(
            [target_dt_idx, target_num_idx], [fc_dt_idx, fc_num_idx]
        ):
            for model in dual_models:
                # skip models which do not support RangeIndex
                if isinstance(target.time_index, pd.RangeIndex):
                    try:
                        # _supports_range_index raises a ValueError if model does not support RangeIndex
                        model._supports_range_index()
                    except ValueError:
                        continue

                # Test models runnability - proper future covariates slicing
                model.fit(target, future_covariates=future_covariates)
                prediction = model.predict(
                    self.forecasting_horizon, future_covariates=future_covariates
                )

                self.assertTrue(len(prediction) == self.forecasting_horizon)

                # Test mismatch in length between exogenous variables and forecasting horizon
                with self.assertRaises(ValueError):
                    model.predict(
                        self.forecasting_horizon,
                        future_covariates=tg.gaussian_timeseries(
                            start=future_covariates.start_time(),
                            length=self.forecasting_horizon - 1,
                        ),
                    )

                # Test mismatch in time-index/length between series and exogenous variables
                with self.assertRaises(ValueError):
                    model.fit(target, future_covariates=target[:-1])
                with self.assertRaises(ValueError):
                    model.fit(target[1:], future_covariates=target[:-1])

    def test_encoders_support(self):
        # test case with pd.DatetimeIndex
        n = 3

        target = self.ts_gaussian[:-3]
        future_covariates = self.ts_gaussian

        add_encoders = {"custom": {"future": [lambda x: x.dayofweek]}}

        # test some models that do not support encoders
        no_support_model_cls = [NaiveMean, Theta]
        for model_cls in no_support_model_cls:
            with pytest.raises(TypeError):
                _ = model_cls(add_encoders=add_encoders)

        # test some models that support encoders
        for model_object in encoder_support_models:
            series = (
                target
                if not isinstance(model_object, VARIMA)
                else target.stack(target.map(np.log))
            )
            # test once with user supplied covariates, and once without
            for fc in [future_covariates, None]:
                model_params = {
                    k: vals
                    for k, vals in copy.deepcopy(model_object.model_params).items()
                }
                model_params["add_encoders"] = add_encoders
                model = model_object.__class__(**model_params)

                # Test models with user supplied covariates
                model.fit(series, future_covariates=fc)

                prediction = model.predict(n, future_covariates=fc)
                self.assertTrue(len(prediction) == n)

                if isinstance(model, TransferableFutureCovariatesLocalForecastingModel):
                    prediction = model.predict(n, series=series, future_covariates=fc)
                    self.assertTrue(len(prediction) == n)

    def test_dummy_series(self):
        values = np.random.uniform(low=-10, high=10, size=100)
        ts = TimeSeries.from_dataframe(pd.DataFrame({"V1": values}))

        varima = VARIMA(trend="t")
        with self.assertRaises(ValueError):
            varima.fit(series=ts)

        autoarima = AutoARIMA(trend="t")
        with self.assertRaises(ValueError):
            autoarima.fit(series=ts)

    def test_forecast_time_index(self):
        # the forecast time index should follow that of the train series

        # integer-index, with step>1
        values = np.random.rand(20)
        idx = pd.RangeIndex(start=10, stop=50, step=2)
        ts = TimeSeries.from_times_and_values(idx, values)

        model = NaiveSeasonal(K=1)
        model.fit(ts)
        pred = model.predict(n=5)
        self.assertTrue(
            all(pred.time_index == pd.RangeIndex(start=50, stop=60, step=2))
        )

        # datetime-index
        ts = tg.constant_timeseries(start=pd.Timestamp("20130101"), length=20, value=1)
        model = NaiveSeasonal(K=1)
        model.fit(ts)
        pred = model.predict(n=5)
        self.assertEqual(pred.start_time(), pd.Timestamp("20130121"))
        self.assertEqual(pred.end_time(), pd.Timestamp("20130125"))

    @pytest.mark.slow
    def test_statsmodels_future_models(self):

        # same tests, but VARIMA requires to work on a multivariate target series
        UNIVARIATE = "univariate"
        MULTIVARIATE = "multivariate"

        params = [
            (ARIMA, {}, UNIVARIATE),
            (VARIMA, {"d": 0}, MULTIVARIATE),
            (VARIMA, {"d": 1}, MULTIVARIATE),
        ]

        for model_cls, kwargs, model_type in params:
            pred_len = 5
            if model_type == MULTIVARIATE:
                series1 = self.ts_ice_heater_train
                series2 = self.ts_ice_heater_val
            else:
                series1 = self.ts_pass_train
                series2 = self.ts_pass_val

            # creating covariates from series + noise
            noise1 = tg.gaussian_timeseries(length=len(series1))
            noise2 = tg.gaussian_timeseries(length=len(series2))

            for _ in range(1, series1.n_components):
                noise1 = noise1.stack(tg.gaussian_timeseries(length=len(series1)))
                noise2 = noise2.stack(tg.gaussian_timeseries(length=len(series2)))

            exog1 = series1 + noise1
            exog2 = series2 + noise2

            exog1_longer = exog1.concatenate(exog1, ignore_time_axis=True)
            exog2_longer = exog2.concatenate(exog2, ignore_time_axis=True)

            # shortening of pred_len so that exog are enough for the training series prediction
            series1 = series1[:-pred_len]
            series2 = series2[:-pred_len]

            # check runnability with different time series
            model = model_cls(**kwargs)
            model.fit(series1)
            pred1 = model.predict(n=pred_len)
            pred2 = model.predict(n=pred_len, series=series2)

            # check probabilistic forecast
            n_samples = 3
            pred1 = model.predict(n=pred_len, num_samples=n_samples)
            pred2 = model.predict(n=pred_len, series=series2, num_samples=n_samples)

            # check that the results with a second custom ts are different from the results given with the training ts
            self.assertFalse(np.array_equal(pred1.values, pred2.values()))

            # check runnability with exogeneous variables
            model = model_cls(**kwargs)
            model.fit(series1, future_covariates=exog1)
            pred1 = model.predict(n=pred_len, future_covariates=exog1)
            pred2 = model.predict(n=pred_len, series=series2, future_covariates=exog2)

            self.assertFalse(np.array_equal(pred1.values(), pred2.values()))

            # check runnability with future covariates with extra time steps in the past compared to the target series
            model = model_cls(**kwargs)
            model.fit(series1, future_covariates=exog1_longer)
            pred1 = model.predict(n=pred_len, future_covariates=exog1_longer)
            pred2 = model.predict(
                n=pred_len, series=series2, future_covariates=exog2_longer
            )

            # check error is raised if model expects covariates but those are not passed when predicting with new data
            with self.assertRaises(ValueError):
                model = model_cls(**kwargs)
                model.fit(series1, future_covariates=exog1)
                model.predict(n=pred_len, series=series2)

            # check error is raised if new future covariates are not wide enough for prediction (on the original series)
            with self.assertRaises(ValueError):
                model = model_cls(**kwargs)
                model.fit(series1, future_covariates=exog1)
                model.predict(n=pred_len, future_covariates=exog1[:-pred_len])

            # check error is raised if new future covariates are not wide enough for prediction (on a new series)
            with self.assertRaises(ValueError):
                model = model_cls(**kwargs)
                model.fit(series1, future_covariates=exog1)
                model.predict(
                    n=pred_len, series=series2, future_covariates=exog2[:-pred_len]
                )
            # and checking the case with insufficient historic future covariates
            with self.assertRaises(ValueError):
                model = model_cls(**kwargs)
                model.fit(series1, future_covariates=exog1)
                model.predict(
                    n=pred_len, series=series2, future_covariates=exog2[pred_len:]
                )

            # verify that we can still forecast the original training series after predicting a new target series
            model = model_cls(**kwargs)
            model.fit(series1, future_covariates=exog1)
            pred1 = model.predict(n=pred_len, future_covariates=exog1)
            model.predict(n=pred_len, series=series2, future_covariates=exog2)
            pred3 = model.predict(n=pred_len, future_covariates=exog1)

            self.assertTrue(np.array_equal(pred1.values(), pred3.values()))
            model.backtest(series1, future_covariates=exog1, start=0.5, retrain=False)

    @patch("typing.Callable", autospec=retrain_func, return_value=True)
    def test_backtest_retrain(
        self,
        patch_retrain_func,
    ):
        """
        Test backtest method with different retrain arguments
        """

        series = self.ts_pass_train

        lr_univ_args = {"lags": [-1, -2, -3]}

        lr_multi_args = {
            "lags": [-1, -2, -3],
            "lags_past_covariates": [-1, -2, -3],
        }
        params = [  # tuple of (model, retrain-able, multivariate, retrain parameter, model type)
            (ExponentialSmoothing(), False, False, "hello", "LocalForecastingModel"),
            (ExponentialSmoothing(), False, False, True, "LocalForecastingModel"),
            (ExponentialSmoothing(), False, False, -2, "LocalForecastingModel"),
            (ExponentialSmoothing(), False, False, 2, "LocalForecastingModel"),
            (
                ExponentialSmoothing(),
                False,
                False,
                patch_retrain_func,
                "LocalForecastingModel",
            ),
            (
                LinearRegressionModel(**lr_univ_args),
                True,
                False,
                True,
                "GlobalForecastingModel",
            ),
            (
                LinearRegressionModel(**lr_univ_args),
                True,
                False,
                2,
                "GlobalForecastingModel",
            ),
            (
                LinearRegressionModel(**lr_univ_args),
                True,
                False,
                patch_retrain_func,
                "GlobalForecastingModel",
            ),
            (
                LinearRegressionModel(**lr_multi_args),
                True,
                True,
                True,
                "GlobalForecastingModel",
            ),
            (
                LinearRegressionModel(**lr_multi_args),
                True,
                True,
                2,
                "GlobalForecastingModel",
            ),
            (
                LinearRegressionModel(**lr_multi_args),
                True,
                True,
                patch_retrain_func,
                "GlobalForecastingModel",
            ),
        ]

        for model_cls, retrainable, multivariate, retrain, model_type in params:

            if (
                not isinstance(retrain, (int, bool, Callable))
                or (isinstance(retrain, int) and retrain < 0)
                or (isinstance(retrain, (Callable)) and (not retrainable))
                or ((retrain != 1) and (not retrainable))
            ):
                with self.assertRaises(ValueError):
                    _ = model_cls.historical_forecasts(series, retrain=retrain)

            else:

                if isinstance(retrain, Mock):
                    # resets patch_retrain_func call_count to 0
                    retrain.call_count = 0
                    retrain.side_effect = [True, False] * (len(series) // 2)

                fit_method_to_patch = f"darts.models.forecasting.forecasting_model.{model_type}._fit_wrapper"
                predict_method_to_patch = f"darts.models.forecasting.forecasting_model.{model_type}._predict_wrapper"

                with patch(fit_method_to_patch) as patch_fit_method:
                    with patch(
                        predict_method_to_patch, side_effect=series
                    ) as patch_predict_method:

                        # Set _fit_called attribute to True, otherwise retrain function is never called
                        model_cls._fit_called = True

                        # run backtest
                        _ = model_cls.historical_forecasts(
                            series,
                            past_covariates=series if multivariate else None,
                            retrain=retrain,
                        )

                        assert patch_predict_method.call_count > 1
                        assert patch_fit_method.call_count > 1

                        if isinstance(retrain, Mock):
                            # check that patch_retrain_func has been called at each iteration
                            assert retrain.call_count > 1

    def test_model_str_call(self):
        model_expected_name_pairs = [
            (ExponentialSmoothing(), "ExponentialSmoothing()"),  # no params changed
            (ARIMA(1, 1, 1), "ARIMA(p=1, q=1)"),  # default value for a param
            (
                Prophet(
                    add_encoders={"cyclic": {"past": ["month"]}}
                ),  # data structure param
                "Prophet(add_encoders={'cyclic': {'past': ['month']}})",
            ),
            (
                TBATS(
                    use_trend=True, use_arma_errors=True, use_box_cox=True
                ),  # params in wrong order
                "TBATS(use_box_cox=True, use_trend=True)",
            ),
        ]
        for model, expected in model_expected_name_pairs:
            self.assertEqual(expected, str(model))

    def test_model_repr_call(self):
        model_expected_name_pairs = [
            (
                ExponentialSmoothing(),
                "ExponentialSmoothing(trend=ModelMode.ADDITIVE, damped=False, seasonal=SeasonalityMode.ADDITIVE, "
                + "seasonal_periods=None, random_state=0)",
            ),  # no params changed
            (
                ARIMA(1, 1, 1),
                "ARIMA(p=1, d=1, q=1, seasonal_order=(0, 0, 0, 0), trend=None, random_state=0, add_encoders=None)",
            ),  # default value for a param
        ]
        for model, expected in model_expected_name_pairs:
            self.assertEqual(expected, repr(model))



import unittest

import numpy as np
import pandas as pd

from darts import TimeSeries
from darts.dataprocessing.transformers.mappers import InvertibleMapper, Mapper
from darts.utils.timeseries_generation import constant_timeseries, linear_timeseries


class MappersTestCase(unittest.TestCase):
    @staticmethod
    def func(x):
        return x + 10

    @staticmethod
    def inverse_func(x):
        return x - 10

    @staticmethod
    def ts_func(ts, x):
        return x - ts.month

    @staticmethod
    def inverse_ts_func(ts, x):
        return x + ts.month

    plus_ten = Mapper(func.__func__)
    plus_ten_invertible = InvertibleMapper(func.__func__, inverse_func.__func__)

    subtract_month = Mapper(ts_func.__func__)
    subtract_month_invertible = InvertibleMapper(
        ts_func.__func__, inverse_ts_func.__func__
    )

    lin_series = linear_timeseries(
        start_value=1,
        length=12,
        freq="MS",
        start=pd.Timestamp("2000-01-01"),
        end_value=12,
    )  # noqa: E501
    zeroes = constant_timeseries(
        value=0.0, length=12, freq="MS", start=pd.Timestamp("2000-01-01")
    )
    tens = constant_timeseries(
        value=10.0, length=12, freq="MS", start=pd.Timestamp("2000-01-01")
    )
    twenties = constant_timeseries(
        value=20.0, length=12, freq="MS", start=pd.Timestamp("2000-01-01")
    )

    def test_mapper(self):

        test_cases = [
            (self.zeroes, self.tens),
            ([self.zeroes, self.tens], [self.tens, self.twenties]),
        ]

        for to_transform, expected_output in test_cases:
            transformed = self.plus_ten.transform(to_transform)
            self.assertEqual(transformed, expected_output)

    def test_invertible_mapper(self):
        test_cases = [(self.zeroes), ([self.zeroes, self.tens])]

        for data in test_cases:
            transformed = self.plus_ten_invertible.transform(data)
            back = self.plus_ten_invertible.inverse_transform(transformed)
            self.assertEqual(back, data)

    def test_mapper_with_timestamp(self):

        test_cases = [
            (self.lin_series, self.zeroes),
            ([self.lin_series, self.lin_series], [self.zeroes, self.zeroes]),
        ]

        for to_transform, expected_output in test_cases:
            transformed = self.subtract_month.transform(to_transform)
            if isinstance(to_transform, list):
                expected_output = [
                    o.with_columns_renamed(o.components[0], t.components[0])
                    for t, o in zip(transformed, expected_output)
                ]
            else:
                expected_output = expected_output.with_columns_renamed(
                    expected_output.components[0], transformed.components[0]
                )
            self.assertEqual(transformed, expected_output)

    def test_invertible_mapper_with_timestamp(self):

        test_cases = [(self.lin_series), ([self.lin_series, self.lin_series])]

        for data in test_cases:
            transformed = self.subtract_month_invertible.transform(data)
            back = self.subtract_month_invertible.inverse_transform(transformed)
            self.assertEqual(back, data)

    def test_invertible_mappers_on_stochastic_series(self):
        vals = np.random.rand(10, 2, 100) + 2
        series = TimeSeries.from_values(vals)

        imapper = InvertibleMapper(np.log, np.exp)
        tr = imapper.transform(series)
        inv_tr = imapper.inverse_transform(tr)

        np.testing.assert_almost_equal(
            series.all_values(copy=False), inv_tr.all_values(copy=False)
        )
